# Copyright (C) 2006-2009 Canonical Ltd

# Authors: Robert Collins <robert.collins@canonical.com>
#          Jelmer Vernooij <jelmer@samba.org>
#          John Carr <john.carr@unrouted.co.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Git-specific subcommands for Bazaar."""

from __future__ import absolute_import

from ...commands import (
    Command,
    display_command,
    )
from ...option import (
    Option,
    )


class cmd_git_import(Command):
    """Import all branches from a git repository.

    """

    takes_args = ["src_location", "dest_location?"]

    takes_options = [
                     Option('colocated', help='Create colocated branches.'),
                     ]

    def _get_colocated_branch(self, target_bzrdir, name):
        from ...errors import NotBranchError
        try:
            return target_bzrdir.open_branch(name=name)
        except NotBranchError:
            return target_bzrdir.create_branch(name=name)

    def _get_nested_branch(self, dest_transport, dest_format, name):
        from ...bzrdir import BzrDir
        from ...errors import NotBranchError
        head_transport = dest_transport.clone(name)
        try:
            head_bzrdir = BzrDir.open_from_transport(head_transport)
        except NotBranchError:
            head_bzrdir = dest_format.initialize_on_transport_ex(
                head_transport, create_prefix=True)[1]
        try:
            return head_bzrdir.open_branch()
        except NotBranchError:
            return head_bzrdir.create_branch()

    def run(self, src_location, dest_location=None, colocated=False):
        import os
        import urllib
        from ... import (
            controldir,
            trace,
            ui,
            urlutils,
            )
        from ...bzrdir import (
            BzrDir,
            )
        from ...errors import (
            BzrCommandError,
            NoRepositoryPresent,
            NotBranchError,
            )
        from . import gettext
        from ...repository import (
            InterRepository,
            Repository,
            )
        from ...transport import get_transport
        from .branch import (
            GitBranch,
            )
        from .refs import (
            ref_to_branch_name,
            )
        from .repository import GitRepository

        dest_format = controldir.ControlDirFormat.get_default_format()

        if dest_location is None:
            dest_location = os.path.basename(src_location.rstrip("/\\"))

        dest_transport = get_transport(dest_location)

        source_repo = Repository.open(src_location)
        if not isinstance(source_repo, GitRepository):
            raise BzrCommandError(gettext("%r is not a git repository") % src_location)
        try:
            target_bzrdir = BzrDir.open_from_transport(dest_transport)
        except NotBranchError:
            target_bzrdir = dest_format.initialize_on_transport_ex(
                dest_transport, shared_repo=True)[1]
        try:
            target_repo = target_bzrdir.find_repository()
        except NoRepositoryPresent:
            target_repo = target_bzrdir.create_repository(shared=True)

        if not target_repo.supports_rich_root():
            raise BzrCommandError(gettext("Target repository doesn't support rich roots"))

        interrepo = InterRepository.get(source_repo, target_repo)
        mapping = source_repo.get_mapping()
        refs = interrepo.fetch()
        refs_dict = refs.as_dict()
        pb = ui.ui_factory.nested_progress_bar()
        try:
            for i, (name, sha) in enumerate(refs_dict.iteritems()):
                try:
                    branch_name = ref_to_branch_name(name)
                except ValueError:
                    # Not a branch, ignore
                    continue
                pb.update(gettext("creating branches"), i, len(refs_dict))
                if getattr(target_bzrdir._format, "colocated_branches", False) and colocated:
                    if name == "HEAD":
                        branch_name = None
                    head_branch = self._get_colocated_branch(target_bzrdir, branch_name)
                else:
                    head_branch = self._get_nested_branch(dest_transport, dest_format, branch_name)
                revid = mapping.revision_id_foreign_to_bzr(sha)
                source_branch = GitBranch(source_repo.bzrdir, source_repo,
                    sha)
                source_branch.head = sha
                if head_branch.last_revision() != revid:
                    head_branch.generate_revision_history(revid)
                source_branch.tags.merge_to(head_branch.tags)
                if not head_branch.get_parent():
                    url = urlutils.join_segment_parameters(
                        source_branch.base, {"ref": urllib.quote(name, '')})
                    head_branch.set_parent(url)
        finally:
            pb.finished()
        trace.note(gettext(
            "Use 'bzr checkout' to create a working tree in "
            "the newly created branches."))


class cmd_git_object(Command):
    """List or display Git objects by SHA.

    Cat a particular object's Git representation if a SHA is specified.
    List all available SHAs otherwise.
    """

    hidden = True

    aliases = ["git-objects", "git-cat"]
    takes_args = ["sha1?"]
    takes_options = [Option('directory',
        short_name='d',
        help='Location of repository.', type=unicode),
        Option('pretty', help='Pretty-print objects.')]
    encoding_type = 'exact'

    @display_command
    def run(self, sha1=None, directory=".", pretty=False):
        from ...errors import (
            BzrCommandError,
            )
        from ...bzrdir import (
            BzrDir,
            )
        from .object_store import (
            get_object_store,
            )
        from . import gettext
        bzrdir, _ = BzrDir.open_containing(directory)
        repo = bzrdir.find_repository()
        object_store = get_object_store(repo)
        object_store.lock_read()
        try:
            if sha1 is not None:
                try:
                    obj = object_store[str(sha1)]
                except KeyError:
                    raise BzrCommandError(gettext("Object not found: %s") % sha1)
                if pretty:
                    text = obj.as_pretty_string()
                else:
                    text = obj.as_raw_string()
                self.outf.write(text)
            else:
                for sha1 in object_store:
                    self.outf.write("%s\n" % sha1)
        finally:
            object_store.unlock()


class cmd_git_refs(Command):
    """Output all of the virtual refs for a repository.

    """

    hidden = True

    takes_args = ["location?"]

    @display_command
    def run(self, location="."):
        from ...bzrdir import (
            BzrDir,
            )
        from .refs import (
            get_refs_container,
            )
        from .object_store import (
            get_object_store,
            )
        bzrdir, _ = BzrDir.open_containing(location)
        repo = bzrdir.find_repository()
        object_store = get_object_store(repo)
        object_store.lock_read()
        try:
            refs = get_refs_container(bzrdir, object_store)
            for k, v in refs.as_dict().iteritems():
                self.outf.write("%s -> %s\n" % (k, v))
        finally:
            object_store.unlock()


class cmd_git_apply(Command):
    """Apply a series of git-am style patches.

    This command will in the future probably be integrated into 
    "bzr pull".
    """

    takes_options = [
        Option('signoff', short_name='s', help='Add a Signed-off-by line.'),
        Option('force',
            help='Apply patches even if tree has uncommitted changes.')
        ]
    takes_args = ["patches*"]

    def _apply_patch(self, wt, f, signoff):
        """Apply a patch.

        :param wt: A Bazaar working tree object.
        :param f: Patch file to read.
        :param signoff: Add Signed-Off-By flag.
        """
        from . import gettext
        from ...errors import BzrCommandError
        from dulwich.patch import git_am_patch_split
        import subprocess
        (c, diff, version) = git_am_patch_split(f)
        # FIXME: Cope with git-specific bits in patch
        # FIXME: Add new files to working tree
        p = subprocess.Popen(["patch", "-p1"], stdin=subprocess.PIPE,
            cwd=wt.basedir)
        p.communicate(diff)
        exitcode = p.wait()
        if exitcode != 0:
            raise BzrCommandError(gettext("error running patch"))
        message = c.message
        if signoff:
            signed_off_by = wt.branch.get_config().username()
            message += "Signed-off-by: %s\n" % signed_off_by.encode('utf-8')
        wt.commit(authors=[c.author], message=message)

    def run(self, patches_list=None, signoff=False, force=False):
        from ...errors import UncommittedChanges
        from ...workingtree import WorkingTree
        if patches_list is None:
            patches_list = []

        tree, _ = WorkingTree.open_containing(".")
        if tree.basis_tree().changes_from(tree).has_changed() and not force:
            raise UncommittedChanges(tree)
        tree.lock_write()
        try:
            for patch in patches_list:
                f = open(patch, 'r')
                try:
                    self._apply_patch(tree, f, signoff=signoff)
                finally:
                    f.close()
        finally:
            tree.unlock()


class cmd_git_push_pristine_tar_deltas(Command):
    """Push pristine tar deltas to a git repository."""

    takes_options = [Option('directory',
        short_name='d',
        help='Location of repository.', type=unicode)]
    takes_args = ['target', 'package']

    def run(self, target, package, directory='.'):
        from ...branch import Branch
        from ...errors import (
            BzrCommandError,
            NoSuchRevision,
            )
        from ...trace import warning
        from ...repository import Repository
        from .object_store import get_object_store
        from .pristine_tar import (
            revision_pristine_tar_data,
            store_git_pristine_tar_data,
            )
        source = Branch.open_containing(directory)[0]
        target_bzr = Repository.open(target)
        target = getattr(target_bzr, '_git', None)
        git_store = get_object_store(source.repository)
        self.add_cleanup(git_store.unlock)
        git_store.lock_read()
        if target is None:
            raise BzrCommandError("Target not a git repository")
        tag_dict = source.tags.get_tag_dict()
        for name, revid in tag_dict.iteritems():
            try:
                rev = source.repository.get_revision(revid)
            except NoSuchRevision:
                continue
            try:
                delta, kind = revision_pristine_tar_data(rev)
            except KeyError:
                continue
            gitid = git_store._lookup_revision_sha1(revid)
            if not (name.startswith('upstream/') or name.startswith('upstream-')):
                warning("Unexpected pristine tar revision tagged %s. Ignoring.",
                     name)
                continue
            upstream_version = name[len("upstream/"):]
            filename = '%s_%s.orig.tar.%s' % (package, upstream_version, kind)
            if not gitid in target:
                warning("base git id %s for %s missing in target repository",
                        gitid, filename)
            store_git_pristine_tar_data(target, filename.encode('utf-8'),
                delta, gitid)



# Copyright (C) 2007 Canonical Ltd
# Copyright (C) 2008-2010 Jelmer Vernooij <jelmer@samba.org>
# Copyright (C) 2008 John Carr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Converters, etc for going between Bazaar and Git ids."""

from __future__ import absolute_import

import base64
import stat

from ... import (
    bencode,
    errors,
    foreign,
    trace,
    )
from ...inventory import (
    ROOT_ID,
    )
from ...foreign import (
    ForeignVcs,
    VcsMappingRegistry,
    ForeignRevision,
    )
from ...revision import (
    NULL_REVISION,
    )
from .errors import (
    NoPushSupport,
    UnknownCommitExtra,
    UnknownMercurialCommitExtra,
    )
from .hg import (
    format_hg_metadata,
    extract_hg_metadata,
    )
from .roundtrip import (
    extract_bzr_metadata,
    inject_bzr_metadata,
    CommitSupplement,
    deserialize_fileid_map,
    serialize_fileid_map,
    )

DEFAULT_FILE_MODE = stat.S_IFREG | 0644
HG_RENAME_SOURCE = "HG:rename-source"
HG_EXTRA = "HG:extra"

# This HG extra is used to indicate the commit that this commit was based on.
HG_EXTRA_AMEND_SOURCE = "amend_source"


def escape_file_id(file_id):
    return file_id.replace('_', '__').replace(' ', '_s').replace('\x0c', '_c')


def unescape_file_id(file_id):
    ret = []
    i = 0
    while i < len(file_id):
        if file_id[i] != '_':
            ret.append(file_id[i])
        else:
            if file_id[i+1] == '_':
                ret.append("_")
            elif file_id[i+1] == 's':
                ret.append(" ")
            elif file_id[i+1] == 'c':
                ret.append("\x0c")
            else:
                raise AssertionError("unknown escape character %s" %
                    file_id[i+1])
            i += 1
        i += 1
    return "".join(ret)


def fix_person_identifier(text):
    if "<" in text and ">" in text:
        if not " <" in text and text.count("<") == 1:
            text = text.replace("<", " <")
        return text
    return "%s <%s>" % (text, text)


def warn_escaped(commit, num_escaped):
    trace.warning("Escaped %d XML-invalid characters in %s. Will be unable "
                  "to regenerate the SHA map.", num_escaped, commit)


def warn_unusual_mode(commit, path, mode):
    trace.mutter("Unusual file mode %o for %s in %s. Storing as revision "
                 "property. ", mode, path, commit)


class BzrGitMapping(foreign.VcsMapping):
    """Class that maps between Git and Bazaar semantics."""
    experimental = False

    BZR_FILE_IDS_FILE = None

    BZR_DUMMY_FILE = None

    def is_special_file(self, filename):
        return (filename in (self.BZR_FILE_IDS_FILE, self.BZR_DUMMY_FILE))

    def __init__(self):
        super(BzrGitMapping, self).__init__(foreign_vcs_git)

    def __eq__(self, other):
        return (type(self) == type(other) and
                self.revid_prefix == other.revid_prefix)

    @classmethod
    def revision_id_foreign_to_bzr(cls, git_rev_id):
        """Convert a git revision id handle to a Bazaar revision id."""
        from dulwich.protocol import ZERO_SHA
        if git_rev_id == ZERO_SHA:
            return NULL_REVISION
        return "%s:%s" % (cls.revid_prefix, git_rev_id)

    @classmethod
    def revision_id_bzr_to_foreign(cls, bzr_rev_id):
        """Convert a Bazaar revision id to a git revision id handle."""
        if not bzr_rev_id.startswith("%s:" % cls.revid_prefix):
            raise errors.InvalidRevisionId(bzr_rev_id, cls)
        return bzr_rev_id[len(cls.revid_prefix)+1:], cls()

    def generate_file_id(self, path):
        # Git paths are just bytestrings
        # We must just hope they are valid UTF-8..
        if path == "":
            return ROOT_ID
        if type(path) is unicode:
            path = path.encode("utf-8")
        return escape_file_id(path)

    def is_control_file(self, path):
        return path in (self.BZR_FILE_IDS_FILE, self.BZR_DUMMY_FILE)

    def parse_file_id(self, file_id):
        if file_id == ROOT_ID:
            return ""
        return unescape_file_id(file_id)

    def revid_as_refname(self, revid):
        import urllib
        return "refs/bzr/%s" % urllib.quote(revid)

    def import_unusual_file_modes(self, rev, unusual_file_modes):
        if unusual_file_modes:
            ret = [(path, unusual_file_modes[path])
                   for path in sorted(unusual_file_modes.keys())]
            rev.properties['file-modes'] = bencode.bencode(ret)

    def export_unusual_file_modes(self, rev):
        try:
            file_modes = rev.properties['file-modes']
        except KeyError:
            return {}
        else:
            return dict([(self.generate_file_id(path), mode) for (path, mode) in bencode.bdecode(file_modes.encode("utf-8"))])

    def _generate_git_svn_metadata(self, rev, encoding):
        try:
            git_svn_id = rev.properties["git-svn-id"]
        except KeyError:
            return ""
        else:
            return "\ngit-svn-id: %s\n" % git_svn_id.encode(encoding)

    def _generate_hg_message_tail(self, rev):
        extra = {}
        renames = []
        branch = 'default'
        for name in rev.properties:
            if name == 'hg:extra:branch':
                branch = rev.properties['hg:extra:branch']
            elif name.startswith('hg:extra'):
                extra[name[len('hg:extra:'):]] = base64.b64decode(
                    rev.properties[name])
            elif name == 'hg:renames':
                renames = bencode.bdecode(base64.b64decode(
                    rev.properties['hg:renames']))
            # TODO: Export other properties as 'bzr:' extras?
        ret = format_hg_metadata(renames, branch, extra)
        assert isinstance(ret, str)
        return ret

    def _extract_git_svn_metadata(self, rev, message):
        lines = message.split("\n")
        if not (lines[-1] == "" and len(lines) >= 2 and lines[-2].startswith("git-svn-id:")):
            return message
        git_svn_id = lines[-2].split(": ", 1)[1]
        rev.properties['git-svn-id'] = git_svn_id
        (url, rev, uuid) = parse_git_svn_id(git_svn_id)
        # FIXME: Convert this to converted-from property somehow..
        ret = "\n".join(lines[:-2])
        assert isinstance(ret, str)
        return ret

    def _extract_hg_metadata(self, rev, message):
        (message, renames, branch, extra) = extract_hg_metadata(message)
        if branch is not None:
            rev.properties['hg:extra:branch'] = branch
        for name, value in extra.iteritems():
            rev.properties['hg:extra:' + name] = base64.b64encode(value)
        if renames:
            rev.properties['hg:renames'] = base64.b64encode(bencode.bencode(
                [(new, old) for (old, new) in renames.iteritems()]))
        return message

    def _extract_bzr_metadata(self, rev, message):
        (message, metadata) = extract_bzr_metadata(message)
        return message, metadata

    def _decode_commit_message(self, rev, message, encoding):
        return message.decode(encoding), CommitSupplement()

    def _encode_commit_message(self, rev, message, encoding):
        return message.encode(encoding)

    def export_fileid_map(self, fileid_map):
        """Export a file id map to a fileid map.

        :param fileid_map: File id map, mapping paths to file ids
        :return: A Git blob object
        """
        from dulwich.objects import Blob
        b = Blob()
        b.set_raw_chunks(serialize_fileid_map(fileid_map))
        return b

    def export_commit(self, rev, tree_sha, parent_lookup, lossy,
                      verifiers):
        """Turn a Bazaar revision in to a Git commit

        :param tree_sha: Tree sha for the commit
        :param parent_lookup: Function for looking up the GIT sha equiv of a
            bzr revision
        :param lossy: Whether to store roundtripping information.
        :param verifiers: Verifiers info
        :return dulwich.objects.Commit represent the revision:
        """
        from dulwich.objects import Commit
        commit = Commit()
        commit.tree = tree_sha
        if not lossy:
            metadata = CommitSupplement()
            metadata.verifiers = verifiers
        else:
            metadata = None
        parents = []
        for p in rev.parent_ids:
            try:
                git_p = parent_lookup(p)
            except KeyError:
                git_p = None
                if metadata is not None:
                    metadata.explicit_parent_ids = rev.parent_ids
            if git_p is not None:
                assert len(git_p) == 40, "unexpected length for %r" % git_p
                parents.append(git_p)
        commit.parents = parents
        try:
            encoding = rev.properties['git-explicit-encoding']
        except KeyError:
            encoding = rev.properties.get('git-implicit-encoding', 'utf-8')
        commit.encoding = rev.properties.get('git-explicit-encoding')
        commit.committer = fix_person_identifier(rev.committer.encode(
            encoding))
        commit.author = fix_person_identifier(
            rev.get_apparent_authors()[0].encode(encoding))
        commit.commit_time = long(rev.timestamp)
        if 'author-timestamp' in rev.properties:
            commit.author_time = long(rev.properties['author-timestamp'])
        else:
            commit.author_time = commit.commit_time
        commit._commit_timezone_neg_utc = "commit-timezone-neg-utc" in rev.properties
        commit.commit_timezone = rev.timezone
        commit._author_timezone_neg_utc = "author-timezone-neg-utc" in rev.properties
        if 'author-timezone' in rev.properties:
            commit.author_timezone = int(rev.properties['author-timezone'])
        else:
            commit.author_timezone = commit.commit_timezone
        commit.message = self._encode_commit_message(rev, rev.message,
            encoding)
        assert type(commit.message) == str
        if metadata is not None:
            try:
                mapping_registry.parse_revision_id(rev.revision_id)
            except errors.InvalidRevisionId:
                metadata.revision_id = rev.revision_id
            mapping_properties = set(
                ['author', 'author-timezone', 'author-timezone-neg-utc',
                 'commit-timezone-neg-utc', 'git-implicit-encoding',
                 'git-explicit-encoding', 'author-timestamp', 'file-modes'])
            for k, v in rev.properties.iteritems():
                if not k in mapping_properties:
                    metadata.properties[k] = v
        if not lossy:
            if self.roundtripping:
                commit.message = inject_bzr_metadata(commit.message, metadata,
                                                     encoding)
            else:
                raise NoPushSupport()
        assert type(commit.message) == str
        if 'git-extra' in rev.properties:
            commit.extra.extend([l.split(' ', 1) for l in rev.properties['git-extra'].splitlines()])
        return commit

    def import_fileid_map(self, blob):
        """Convert a git file id map blob.

        :param blob: Git blob object with fileid map
        :return: Dictionary mapping paths to file ids
        """
        return deserialize_fileid_map(blob.data)

    def import_commit(self, commit, lookup_parent_revid):
        """Convert a git commit to a bzr revision.

        :return: a `bzrlib.revision.Revision` object, foreign revid and a
            testament sha1
        """
        if commit is None:
            raise AssertionError("Commit object can't be None")
        rev = ForeignRevision(commit.id, self,
                self.revision_id_foreign_to_bzr(commit.id))
        rev.git_metadata = None
        def decode_using_encoding(rev, commit, encoding):
            rev.committer = str(commit.committer).decode(encoding)
            if commit.committer != commit.author:
                rev.properties['author'] = str(commit.author).decode(encoding)
            rev.message, rev.git_metadata = self._decode_commit_message(
                rev, commit.message, encoding)
        if commit.encoding is not None:
            rev.properties['git-explicit-encoding'] = commit.encoding
            decode_using_encoding(rev, commit, commit.encoding)
        else:
            for encoding in ('utf-8', 'latin1'):
                try:
                    decode_using_encoding(rev, commit, encoding)
                except UnicodeDecodeError:
                    pass
                else:
                    if encoding != 'utf-8':
                        rev.properties['git-implicit-encoding'] = encoding
                    break
        if commit.commit_time != commit.author_time:
            rev.properties['author-timestamp'] = str(commit.author_time)
        if commit.commit_timezone != commit.author_timezone:
            rev.properties['author-timezone'] = "%d" % commit.author_timezone
        if commit._author_timezone_neg_utc:
            rev.properties['author-timezone-neg-utc'] = ""
        if commit._commit_timezone_neg_utc:
            rev.properties['commit-timezone-neg-utc'] = ""
        rev.timestamp = commit.commit_time
        rev.timezone = commit.commit_timezone
        rev.parent_ids = None
        if rev.git_metadata is not None:
            md = rev.git_metadata
            roundtrip_revid = md.revision_id
            if md.explicit_parent_ids:
                rev.parent_ids = md.explicit_parent_ids
            rev.properties.update(md.properties)
            verifiers = md.verifiers
        else:
            roundtrip_revid = None
            verifiers = {}
        if rev.parent_ids is None:
            rev.parent_ids = tuple([lookup_parent_revid(p) for p in commit.parents])
        unknown_extra_fields = []
        extra_lines = []
        for k, v in commit.extra:
            if k == HG_RENAME_SOURCE:
                extra_lines.append(k + ' ' + v + '\n')
            elif k == HG_EXTRA:
                hgk, hgv = v.split(':', 1)
                if hgk not in (HG_EXTRA_AMEND_SOURCE, ):
                    raise UnknownMercurialCommitExtra(commit, hgk)
                extra_lines.append(k + ' ' + v + '\n')
            else:
                unknown_extra_fields.append(k)
        if unknown_extra_fields:
            raise UnknownCommitExtra(commit, unknown_extra_fields)
        if extra_lines:
            rev.properties['git-extra'] = ''.join(extra_lines)
        return rev, roundtrip_revid, verifiers

    def get_fileid_map(self, lookup_object, tree_sha):
        """Obtain a fileid map for a particular tree.

        :param lookup_object: Function for looking up an object
        :param tree_sha: SHA of the root tree
        :return: GitFileIdMap instance
        """
        try:
            file_id_map_sha = lookup_object(tree_sha)[self.BZR_FILE_IDS_FILE][1]
        except KeyError:
            file_ids = {}
        else:
            file_ids = self.import_fileid_map(lookup_object(file_id_map_sha))
        return GitFileIdMap(file_ids, self)


class BzrGitMappingv1(BzrGitMapping):
    revid_prefix = 'git-v1'
    experimental = False

    def __str__(self):
        return self.revid_prefix


class BzrGitMappingExperimental(BzrGitMappingv1):
    revid_prefix = 'git-experimental'
    experimental = True
    roundtripping = True

    BZR_FILE_IDS_FILE = '.bzrfileids'

    BZR_DUMMY_FILE = '.bzrdummy'

    def _decode_commit_message(self, rev, message, encoding):
        message = self._extract_hg_metadata(rev, message)
        message = self._extract_git_svn_metadata(rev, message)
        message, metadata = self._extract_bzr_metadata(rev, message)
        return message.decode(encoding), metadata

    def _encode_commit_message(self, rev, message, encoding):
        ret = message.encode(encoding)
        ret += self._generate_hg_message_tail(rev)
        ret += self._generate_git_svn_metadata(rev, encoding)
        return ret

    def import_commit(self, commit, lookup_parent_revid):
        rev, roundtrip_revid, verifiers = super(BzrGitMappingExperimental, self).import_commit(commit, lookup_parent_revid)
        rev.properties['converted_revision'] = "git %s\n" % commit.id
        return rev, roundtrip_revid, verifiers


class GitMappingRegistry(VcsMappingRegistry):
    """Registry with available git mappings."""

    def revision_id_bzr_to_foreign(self, bzr_revid):
        if bzr_revid == NULL_REVISION:
            from dulwich.protocol import ZERO_SHA
            return ZERO_SHA, None
        if not bzr_revid.startswith("git-"):
            raise errors.InvalidRevisionId(bzr_revid, None)
        (mapping_version, git_sha) = bzr_revid.split(":", 1)
        mapping = self.get(mapping_version)
        return mapping.revision_id_bzr_to_foreign(bzr_revid)

    parse_revision_id = revision_id_bzr_to_foreign


mapping_registry = GitMappingRegistry()
mapping_registry.register_lazy('git-v1', "bzrlib.plugins.git.mapping",
    "BzrGitMappingv1")
mapping_registry.register_lazy('git-experimental',
    "bzrlib.plugins.git.mapping", "BzrGitMappingExperimental")
# Uncomment the next line to enable the experimental bzr-git mappings.
# This will make sure all bzr metadata is pushed into git, allowing for
# full roundtripping later.
# NOTE: THIS IS EXPERIMENTAL. IT MAY EAT YOUR DATA OR CORRUPT
# YOUR BZR OR GIT REPOSITORIES. USE WITH CARE.
#mapping_registry.set_default('git-experimental')
mapping_registry.set_default('git-v1')


class ForeignGit(ForeignVcs):
    """The Git Stupid Content Tracker"""

    @property
    def branch_format(self):
        from .branch import GitBranchFormat
        return GitBranchFormat()

    @property
    def repository_format(self):
        from .repository import GitRepositoryFormat
        return GitRepositoryFormat()

    def __init__(self):
        super(ForeignGit, self).__init__(mapping_registry)
        self.abbreviation = "git"

    @classmethod
    def serialize_foreign_revid(self, foreign_revid):
        return foreign_revid

    @classmethod
    def show_foreign_revid(cls, foreign_revid):
        return { "git commit": foreign_revid }


foreign_vcs_git = ForeignGit()
default_mapping = mapping_registry.get_default()()


def symlink_to_blob(symlink_target):
    from dulwich.objects import Blob
    blob = Blob()
    if type(symlink_target) == unicode:
        symlink_target = symlink_target.encode('utf-8')
    blob.data = symlink_target
    return blob


def mode_is_executable(mode):
    """Check if mode should be considered executable."""
    return bool(mode & 0111)


def mode_kind(mode):
    """Determine the Bazaar inventory kind based on Unix file mode."""
    if mode is None:
        return None
    entry_kind = (mode & 0700000) / 0100000
    if entry_kind == 0:
        return 'directory'
    elif entry_kind == 1:
        file_kind = (mode & 070000) / 010000
        if file_kind == 0:
            return 'file'
        elif file_kind == 2:
            return 'symlink'
        elif file_kind == 6:
            return 'tree-reference'
        else:
            raise AssertionError(
                "Unknown file kind %d, perms=%o." % (file_kind, mode,))
    else:
        raise AssertionError(
            "Unknown kind, perms=%r." % (mode,))


def object_mode(kind, executable):
    if kind == 'directory':
        return stat.S_IFDIR
    elif kind == 'symlink':
        mode = stat.S_IFLNK
        if executable:
            mode |= 0111
        return mode
    elif kind == 'file':
        mode = stat.S_IFREG | 0644
        if executable:
            mode |= 0111
        return mode
    elif kind == 'tree-reference':
        from dulwich.objects import S_IFGITLINK
        return S_IFGITLINK
    else:
        raise AssertionError


def entry_mode(entry):
    """Determine the git file mode for an inventory entry."""
    return object_mode(entry.kind, entry.executable)


def directory_to_tree(children, lookup_ie_sha1, unusual_modes, empty_file_name,
                      allow_empty=False):
    """Create a Git Tree object from a Bazaar directory.

    :param children: Children inventory entries
    :param lookup_ie_sha1: Lookup the Git SHA1 for a inventory entry
    :param unusual_modes: Dictionary with unusual file modes by file ids
    :param empty_file_name: Name to use for dummy files in empty directories,
        None to ignore empty directories.
    """
    from dulwich.objects import Blob, Tree
    tree = Tree()
    for name, value in children.iteritems():
        ie = children[name]
        try:
            mode = unusual_modes[ie.file_id]
        except KeyError:
            mode = entry_mode(ie)
        hexsha = lookup_ie_sha1(ie)
        if hexsha is not None:
            tree.add(name.encode("utf-8"), mode, hexsha)
    if not allow_empty and len(tree) == 0:
        # Only the root can be an empty tree
        if empty_file_name is not None:
            tree.add(empty_file_name, stat.S_IFREG | 0644, Blob().id)
        else:
            return None
    return tree


def extract_unusual_modes(rev):
    try:
        foreign_revid, mapping = mapping_registry.parse_revision_id(
            rev.revision_id)
    except errors.InvalidRevisionId:
        return {}
    else:
        return mapping.export_unusual_file_modes(rev)


def parse_git_svn_id(text):
    (head, uuid) = text.rsplit(" ", 1)
    (full_url, rev) = head.rsplit("@", 1)
    return (full_url, int(rev), uuid)


class GitFileIdMap(object):

    def __init__(self, file_ids, mapping):
        self.file_ids = file_ids
        self.paths = None
        self.mapping = mapping

    def all_file_ids(self):
        return self.file_ids.values()

    def set_file_id(self, path, file_id):
        assert type(path) is str
        assert type(file_id) is str
        self.file_ids[path] = file_id

    def lookup_file_id(self, path):
        assert type(path) is str
        try:
            file_id = self.file_ids[path]
        except KeyError:
            file_id = self.mapping.generate_file_id(path)
        assert type(file_id) is str
        return file_id

    def lookup_path(self, file_id):
        if self.paths is None:
            self.paths = {}
            for k, v in self.file_ids.iteritems():
                self.paths[v] = k
        try:
            path = self.paths[file_id]
        except KeyError:
            return self.mapping.parse_file_id(file_id)
        else:
            assert type(path) is str
            return path

    def copy(self):
        return self.__class__(dict(self.file_ids), self.mapping)



# Copyright (C) 2011 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Tests for the unpeel map."""

from cStringIO import StringIO

from ....tests import (
    TestCaseWithTransport,
    )

from ..unpeel_map import (
    UnpeelMap,
    )


class TestUnpeelMap(TestCaseWithTransport):

    def test_new(self):
        m = UnpeelMap()
        self.assertIs(None, m.peel_tag("ab"* 20))

    def test_load(self):
        f = StringIO(
            "unpeel map version 1\n"
            "0123456789012345678901234567890123456789: aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\n")
        m = UnpeelMap()
        m.load(f)
        self.assertEquals("0123456789012345678901234567890123456789",
            m.peel_tag("aa"*20))

    def test_update(self):
        m = UnpeelMap()
        m.update({
           "0123456789012345678901234567890123456789": set(["aa" * 20]),
           })
        self.assertEquals("0123456789012345678901234567890123456789",
            m.peel_tag("aa"*20))



# Copyright (C) 2010 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Tests for bzr-git's object store."""


from dulwich.objects import Blob
from dulwich.tests.test_object_store import PackBasedObjectStoreTests
from dulwich.tests.utils import make_object

from ....tests import TestCaseWithTransport

from ..transportgit import TransportObjectStore


class TransportObjectStoreTests(PackBasedObjectStoreTests, TestCaseWithTransport):

    def setUp(self):
        TestCaseWithTransport.setUp(self)
        self.store = TransportObjectStore.init(self.get_transport())

    def tearDown(self):
        PackBasedObjectStoreTests.tearDown(self)
        TestCaseWithTransport.tearDown(self)

    def test_remembers_packs(self):
        self.store.add_object(make_object(Blob, data="data"))
        self.assertEqual(0, len(self.store.packs))
        self.store.pack_loose_objects()
        self.assertEqual(1, len(self.store.packs))

        # Packing a second object creates a second pack.
        self.store.add_object(make_object(Blob, data="more data"))
        self.store.pack_loose_objects()
        self.assertEqual(2, len(self.store.packs))

        # If we reopen the store, it reloads both packs.
        restore = TransportObjectStore(self.get_transport())
        self.assertEqual(2, len(restore.packs))


# FIXME: Unfortunately RefsContainerTests requires on a specific set of refs existing.

# class TransportRefContainerTests(RefsContainerTests, TestCaseWithTransport):
#
#    def setUp(self):
#        TestCaseWithTransport.setUp(self)
#        self._refs = TransportRefsContainer(self.get_transport())




# Copyright (C) 2009 Jelmer Vernooij <jelmer@samba.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""Custom revision specifier for Subversion."""

from __future__ import absolute_import

# Please note that imports are delayed as much as possible here since
# if DWIM revspecs are supported this module is imported by __init__.py.

from ... import version_info as bzrlib_version
from ...errors import (
    InvalidRevisionId,
    InvalidRevisionSpec,
    )
from ...revision import (
    NULL_REVISION,
)
from ...revisionspec import (
    RevisionInfo,
    RevisionSpec,
    )


def valid_git_sha1(hex):
    """Check if `hex` is a validly formatted Git SHA1.

    :param hex: Hex string to validate
    :return: Boolean
    """
    import binascii
    try:
        binascii.unhexlify(hex)
    except TypeError:
        return False
    else:
        return True


class RevisionSpec_git(RevisionSpec):
    """Selects a revision using a Subversion revision number."""

    help_txt = """Selects a revision using a Git revision sha1.
    """

    prefix = 'git:'
    wants_revision_history = False

    def _lookup_git_sha1(self, branch, sha1):
        from .errors import (
            GitSmartRemoteNotSupported,
            )
        from .mapping import (
            default_mapping,
            )

        bzr_revid = getattr(branch.repository, "lookup_foreign_revision_id",
                              default_mapping.revision_id_foreign_to_bzr)(sha1)
        try:
            if branch.repository.has_revision(bzr_revid):
                return RevisionInfo.from_revision_id(branch, bzr_revid)
        except GitSmartRemoteNotSupported:
            return RevisionInfo(branch, None, bzr_revid)
        raise InvalidRevisionSpec(self.user_spec, branch)

    def __nonzero__(self):
        # The default implementation uses branch.repository.has_revision()
        if self.rev_id is None:
            return False
        if self.rev_id == NULL_REVISION:
            return False
        return True

    def _find_short_git_sha1(self, branch, sha1):
        from .mapping import (
            ForeignGit,
            mapping_registry,
            )
        parse_revid = getattr(branch.repository, "lookup_bzr_revision_id",
                              mapping_registry.parse_revision_id)
        branch.repository.lock_read()
        try:
            graph = branch.repository.get_graph()
            for revid, _ in graph.iter_ancestry([branch.last_revision()]):
                if revid == NULL_REVISION:
                    continue
                try:
                    foreign_revid, mapping = parse_revid(revid)
                except InvalidRevisionId:
                    continue
                if not isinstance(mapping.vcs, ForeignGit):
                    continue
                if foreign_revid.startswith(sha1):
                    return RevisionInfo.from_revision_id(branch, revid)
            raise InvalidRevisionSpec(self.user_spec, branch)
        finally:
            branch.repository.unlock()

    def _match_on(self, branch, revs):
        loc = self.spec.find(':')
        git_sha1 = self.spec[loc+1:].encode("utf-8")
        if len(git_sha1) > 40 or not valid_git_sha1(git_sha1):
            raise InvalidRevisionSpec(self.user_spec, branch)
        from . import (
            lazy_check_versions,
            )
        lazy_check_versions()
        if len(git_sha1) == 40:
            return self._lookup_git_sha1(branch, git_sha1)
        else:
            return self._find_short_git_sha1(branch, git_sha1)

    def needs_branch(self):
        return True

    def get_branch(self):
        return None



# Copyright (C) 2007 Canonical Ltd
# Copyright (C) 2010 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""An adapter between a Git control dir and a Bazaar ControlDir."""

from __future__ import absolute_import

import urllib

from ... import (
    errors as bzr_errors,
    trace,
    osutils,
    revision as _mod_revision,
    urlutils,
    )
from ...bzrdir import CreateRepository
from ...transport import do_catching_redirections

from ...controldir import (
    ControlDir,
    ControlDirFormat,
    format_registry,
    )


class GitDirConfig(object):

    def get_default_stack_on(self):
        return None

    def set_default_stack_on(self, value):
        raise bzr_errors.BzrError("Cannot set configuration")


class GitControlDirFormat(ControlDirFormat):

    colocated_branches = True
    fixed_components = True

    def __eq__(self, other):
        return type(self) == type(other)

    def is_supported(self):
        return True

    def network_name(self):
        return "git"


class GitDir(ControlDir):
    """An adapter to the '.git' dir used by git."""

    def is_supported(self):
        return True

    def can_convert_format(self):
        return False

    def break_lock(self):
        pass

    def cloning_metadir(self, stacked=False):
        return format_registry.make_bzrdir("default")

    def checkout_metadir(self, stacked=False):
        return format_registry.make_bzrdir("default")

    def _get_default_ref(self):
        return "HEAD"

    def _get_selected_ref(self, branch, ref=None):
        if ref is not None and branch is not None:
            raise bzr_errors.BzrError("can't specify both ref and branch")
        if ref is not None:
            return ref
        segment_parameters = getattr(
            self.user_transport, "get_segment_parameters", lambda: {})()
        ref = segment_parameters.get("ref")
        if ref is not None:
            return urlutils.unescape(ref)
        if branch is None and getattr(self, "_get_selected_branch", False):
            branch = self._get_selected_branch()
        if branch is not None:
            from .refs import branch_name_to_ref
            return branch_name_to_ref(branch)
        return self._get_default_ref()

    def get_config(self):
        return GitDirConfig()

    def _available_backup_name(self, base):
        return osutils.available_backup_name(base, self.root_transport.has)

    def sprout(self, url, revision_id=None, force_new_repo=False,
               recurse='down', possible_transports=None,
               accelerator_tree=None, hardlink=False, stacked=False,
               source_branch=None, create_tree_if_local=True):
        from ...repository import InterRepository
        from ...transport.local import LocalTransport
        from ...transport import get_transport
        target_transport = get_transport(url, possible_transports)
        target_transport.ensure_base()
        cloning_format = self.cloning_metadir()
        # Create/update the result branch
        result = cloning_format.initialize_on_transport(target_transport)
        source_branch = self.open_branch()
        source_repository = self.find_repository()
        try:
            result_repo = result.find_repository()
        except bzr_errors.NoRepositoryPresent:
            result_repo = result.create_repository()
            target_is_empty = True
        else:
            target_is_empty = None # Unknown
        if stacked:
            raise bzr_errors.IncompatibleRepositories(source_repository, result_repo)
        interrepo = InterRepository.get(source_repository, result_repo)

        if revision_id is not None:
            determine_wants = interrepo.get_determine_wants_revids(
                [revision_id], include_tags=False)
        else:
            determine_wants = interrepo.determine_wants_all
        interrepo.fetch_objects(determine_wants=determine_wants,
            mapping=source_branch.mapping)
        result_branch = source_branch.sprout(result,
            revision_id=revision_id, repository=result_repo)
        if (create_tree_if_local
            and isinstance(target_transport, LocalTransport)
            and (result_repo is None or result_repo.make_working_trees())):
            wt = result.create_workingtree(accelerator_tree=accelerator_tree,
                hardlink=hardlink, from_branch=result_branch)
            wt.lock_write()
            try:
                if wt.path2id('') is None:
                    try:
                        wt.set_root_id(self.open_workingtree.get_root_id())
                    except bzr_errors.NoWorkingTree:
                        pass
            finally:
                wt.unlock()
        return result

    def clone_on_transport(self, transport, revision_id=None,
        force_new_repo=False, preserve_stacking=False, stacked_on=None,
        create_prefix=False, use_existing_dir=True, no_tree=False):
        """See ControlDir.clone_on_transport."""
        from ...repository import InterRepository
        from .mapping import default_mapping
        if no_tree:
            format = BareLocalGitControlDirFormat()
        else:
            format = LocalGitControlDirFormat()
        (target_repo, target_controldir, stacking, repo_policy) = format.initialize_on_transport_ex(transport, use_existing_dir=use_existing_dir, create_prefix=create_prefix, force_new_repo=force_new_repo)
        target_git_repo = target_repo._git
        source_repo = self.open_repository()
        source_git_repo = source_repo._git
        interrepo = InterRepository.get(source_repo, target_repo)
        if revision_id is not None:
            determine_wants = interrepo.get_determine_wants_revids([revision_id], include_tags=True)
        else:
            determine_wants = interrepo.determine_wants_all
        (pack_hint, _, refs) = interrepo.fetch_objects(determine_wants,
            mapping=default_mapping)
        for name, val in refs.iteritems():
            target_git_repo.refs[name] = val
        return self.__class__(transport, target_git_repo, format)

    def find_repository(self):
        """Find the repository that should be used.

        This does not require a branch as we use it to find the repo for
        new branches as well as to hook existing branches up to their
        repository.
        """
        return self.open_repository()

    def get_refs_container(self):
        """Retrieve the refs container.
        """
        raise NotImplementedError(self.get_refs_container)


class LocalGitControlDirFormat(GitControlDirFormat):
    """The .git directory control format."""

    bare = False

    @classmethod
    def _known_formats(self):
        return set([LocalGitControlDirFormat()])

    @property
    def repository_format(self):
        from .repository import GitRepositoryFormat
        return GitRepositoryFormat()

    def get_branch_format(self):
        from .branch import GitBranchFormat
        return GitBranchFormat()

    def open(self, transport, _found=None):
        """Open this directory.

        """
        from .transportgit import TransportRepo
        gitrepo = TransportRepo(transport, self.bare,
                refs_text=getattr(self, "_refs_text", None))
        return LocalGitDir(transport, gitrepo, self)

    def get_format_description(self):
        return "Local Git Repository"

    def initialize_on_transport(self, transport):
        from .transportgit import TransportRepo
        repo = TransportRepo.init(transport, bare=self.bare)
        del repo.refs["HEAD"]
        return self.open(transport)

    def initialize_on_transport_ex(self, transport, use_existing_dir=False,
        create_prefix=False, force_new_repo=False, stacked_on=None,
        stack_on_pwd=None, repo_format_name=None, make_working_trees=None,
        shared_repo=False, vfs_only=False):
        def make_directory(transport):
            transport.mkdir('.')
            return transport
        def redirected(transport, e, redirection_notice):
            trace.note(redirection_notice)
            return transport._redirected_to(e.source, e.target)
        try:
            transport = do_catching_redirections(make_directory, transport,
                redirected)
        except bzr_errors.FileExists:
            if not use_existing_dir:
                raise
        except bzr_errors.NoSuchFile:
            if not create_prefix:
                raise
            transport.create_prefix()
        controldir = self.initialize_on_transport(transport)
        repository = controldir.open_repository()
        repository.lock_write()
        return (repository, controldir, False, CreateRepository(controldir))

    def is_supported(self):
        return True

    def supports_transport(self, transport):
        try:
            external_url = transport.external_url()
        except bzr_errors.InProcessTransport:
            raise bzr_errors.NotBranchError(path=transport.base)
        return (external_url.startswith("http:") or
                external_url.startswith("https:") or
                external_url.startswith("file:"))


class BareLocalGitControlDirFormat(LocalGitControlDirFormat):

    bare = True
    supports_workingtrees = False

    def get_format_description(self):
        return "Local Git Repository (bare)"


class LocalGitDir(GitDir):
    """An adapter to the '.git' dir used by git."""

    def _get_gitrepository_class(self):
        from .repository import LocalGitRepository
        return LocalGitRepository

    def __repr__(self):
        return "<%s at %r>" % (
            self.__class__.__name__, self.root_transport.base)

    _gitrepository_class = property(_get_gitrepository_class)

    @property
    def user_transport(self):
        return self.root_transport

    @property
    def control_transport(self):
        return self.transport

    def __init__(self, transport, gitrepo, format):
        self._format = format
        self.root_transport = transport
        self._mode_check_done = False
        self._git = gitrepo
        if gitrepo.bare:
            self.transport = transport
        else:
            self.transport = transport.clone('.git')
        self._mode_check_done = None

    def is_control_filename(self, filename):
        return (filename == '.git' or
                filename.startswith('.git/') or
                filename.startswith('.git\\'))

    def _get_symref(self, ref):
        from dulwich.repo import SYMREF
        refcontents = self._git.refs.read_ref(ref)
        if refcontents is None: # no such ref
            return None
        if refcontents.startswith(SYMREF):
            return refcontents[len(SYMREF):].rstrip("\n")
        return None

    def set_branch_reference(self, target, name=None):
        if self.control_transport.base != target.bzrdir.control_transport.base:
            raise bzr_errors.IncompatibleFormat(target._format, self._format)
        ref = self._get_selected_ref(name)
        self._git.refs.set_symbolic_ref(ref, target.ref)

    def get_branch_reference(self, name=None):
        ref = self._get_selected_ref(name)
        target_ref = self._get_symref(ref)
        if target_ref is not None:
            return urlutils.join_segment_parameters(
                self.user_url.rstrip("/"), {"ref": urllib.quote(target_ref, '')})
        return None

    def find_branch_format(self, name=None):
        from .branch import (
            GitBranchFormat,
            GitSymrefBranchFormat,
            )
        ref = self._get_selected_ref(name)
        if self._get_symref(ref) is not None:
            return GitSymrefBranchFormat()
        else:
            return GitBranchFormat()

    def get_branch_transport(self, branch_format, name=None):
        if branch_format is None:
            return self.transport
        if isinstance(branch_format, LocalGitControlDirFormat):
            return self.transport
        raise bzr_errors.IncompatibleFormat(branch_format, self._format)

    def get_repository_transport(self, format):
        if format is None:
            return self.transport
        if isinstance(format, LocalGitControlDirFormat):
            return self.transport
        raise bzr_errors.IncompatibleFormat(format, self._format)

    def get_workingtree_transport(self, format):
        if format is None:
            return self.transport
        if isinstance(format, LocalGitControlDirFormat):
            return self.transport
        raise bzr_errors.IncompatibleFormat(format, self._format)

    def open_branch(self, name=None, unsupported=False, ignore_fallbacks=None,
            ref=None, possible_transports=None):
        """'create' a branch for this dir."""
        repo = self.open_repository()
        from .branch import LocalGitBranch
        ref = self._get_selected_ref(name, ref)
        ref_chain, sha = self._git.refs.follow(ref)
        if sha is None:
            raise bzr_errors.NotBranchError(self.root_transport.base,
                    bzrdir=self)
        return LocalGitBranch(self, repo, ref)

    def destroy_branch(self, name=None):
        refname = self._get_selected_ref(name)
        try:
            del self._git.refs[refname]
        except KeyError:
            raise bzr_errors.NotBranchError(self.root_transport.base,
                    bzrdir=self)

    def destroy_repository(self):
        raise bzr_errors.UnsupportedOperation(self.destroy_repository, self)

    def destroy_workingtree(self):
        wt = self.open_workingtree(recommend_upgrade=False)
        repository = wt.branch.repository
        empty = repository.revision_tree(_mod_revision.NULL_REVISION)
        # We ignore the conflicts returned by wt.revert since we're about to
        # delete the wt metadata anyway, all that should be left here are
        # detritus. But see bug #634470 about subtree .bzr dirs.
        conflicts = wt.revert(old_tree=empty)
        self.destroy_workingtree_metadata()

    def destroy_workingtree_metadata(self):
        self.transport.delete('index')

    def needs_format_conversion(self, format=None):
        return not isinstance(self._format, format.__class__)

    def list_branches(self):
        return self.get_branches().values()

    def get_branches(self):
        from .refs import ref_to_branch_name
        ret = {}
        for ref in self._git.refs.keys():
            try:
                branch_name = ref_to_branch_name(ref)
            except ValueError:
                continue
            except UnicodeDecodeError:
                trace.warning("Ignoring branch %r with unicode error ref", ref)
                continue
            ret[branch_name] = self.open_branch(ref=ref)
        return ret

    def open_repository(self):
        """'open' a repository for this dir."""
        return self._gitrepository_class(self)

    def open_workingtree(self, recommend_upgrade=True, unsupported=False):
        if not self._git.bare:
            from dulwich.errors import NoIndexPresent
            repo = self.open_repository()
            try:
                index = repo._git.open_index()
            except NoIndexPresent:
                pass
            else:
                from .workingtree import GitWorkingTree
                try:
                    branch = self.open_branch()
                except bzr_errors.NotBranchError:
                    pass
                else:
                    return GitWorkingTree(self, repo, branch, index)
        loc = urlutils.unescape_for_display(self.root_transport.base, 'ascii')
        raise bzr_errors.NoWorkingTree(loc)

    def create_repository(self, shared=False):
        from .repository import GitRepositoryFormat
        if shared:
            raise bzr_errors.IncompatibleFormat(GitRepositoryFormat(), self._format)
        return self.open_repository()

    def create_branch(self, name=None, repository=None,
                      append_revisions_only=None, ref=None):
        refname = self._get_selected_ref(name, ref)
        from dulwich.protocol import ZERO_SHA
        if refname in self._git.refs:
            raise bzr_errors.AlreadyBranchError(self.user_url)
        self._git.refs[refname] = ZERO_SHA
        branch = self.open_branch(name)
        if append_revisions_only:
            branch.set_append_revisions_only(append_revisions_only)
        return branch

    def backup_bzrdir(self):
        if not self._git.bare:
            self.root_transport.copy_tree(".git", ".git.backup")
            return (self.root_transport.abspath(".git"),
                    self.root_transport.abspath(".git.backup"))
        else:
            basename = urlutils.basename(self.root_transport.base)
            parent = self.root_transport.clone('..')
            parent.copy_tree(basename, basename + ".backup")

    def create_workingtree(self, revision_id=None, from_branch=None,
        accelerator_tree=None, hardlink=False):
        if self._git.bare:
            raise bzr_errors.UnsupportedOperation(self.create_workingtree, self)
        from dulwich.index import write_index
        from dulwich.pack import SHA1Writer
        f = open(self.transport.local_abspath("index"), 'w+')
        try:
            f = SHA1Writer(f)
            write_index(f, [])
        finally:
            f.close()
        return self.open_workingtree()

    def _find_or_create_repository(self, force_new_repo=None):
        return self.create_repository(shared=False)

    def _find_creation_modes(self):
        """Determine the appropriate modes for files and directories.

        They're always set to be consistent with the base directory,
        assuming that this transport allows setting modes.
        """
        # TODO: Do we need or want an option (maybe a config setting) to turn
        # this off or override it for particular locations? -- mbp 20080512
        if self._mode_check_done:
            return
        self._mode_check_done = True
        try:
            st = self.transport.stat('.')
        except bzr_errors.TransportNotPossible:
            self._dir_mode = None
            self._file_mode = None
        else:
            # Check the directory mode, but also make sure the created
            # directories and files are read-write for this user. This is
            # mostly a workaround for filesystems which lie about being able to
            # write to a directory (cygwin & win32)
            if (st.st_mode & 07777 == 00000):
                # FTP allows stat but does not return dir/file modes
                self._dir_mode = None
                self._file_mode = None
            else:
                self._dir_mode = (st.st_mode & 07777) | 00700
                # Remove the sticky and execute bits for files
                self._file_mode = self._dir_mode & ~07111

    def _get_file_mode(self):
        """Return Unix mode for newly created files, or None.
        """
        if not self._mode_check_done:
            self._find_creation_modes()
        return self._file_mode

    def _get_dir_mode(self):
        """Return Unix mode for newly created directories, or None.
        """
        if not self._mode_check_done:
            self._find_creation_modes()
        return self._dir_mode

    def get_refs_container(self):
        return self._git.refs

    def get_peeled(self, ref):
        return self._git.get_peeled(ref)



# Copyright (C) 2012 Jelmer Vernooij <jelmer@samba.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""Directory service for gitorious."""

from __future__ import absolute_import

from __future__ import absolute_import
from ... import transport

transport.register_urlparse_netloc_protocol('github')

class GitHubDirectory(object):

    def look_up(self, name, url):
        """See DirectoryService.look_up"""
        return "git+ssh://git@github.com/" + name



# Copyright (C) 2010 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Test the git revision specifiers."""

from ....tests import TestCase

from ..revspec import (
    valid_git_sha1,
    )

class Sha1ValidTests(TestCase):

    def test_invalid(self):
        self.assertFalse(valid_git_sha1("git-v1:abcde"))

    def test_valid(self):
        self.assertTrue(valid_git_sha1("aabbccddee"))



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Test the GitDir class"""

from dulwich.repo import Repo as GitRepo
import os

from .... import (
    bzrdir,
    errors,
    urlutils,
    )
from ....tests import TestSkipped

from .. import (
    dir,
    tests,
    workingtree,
    )


class TestGitDir(tests.TestCaseInTempDir):

    def test_get_head_branch_reference(self):
        GitRepo.init(".")

        gd = bzrdir.BzrDir.open('.')
        self.assertEquals(
            "%s,ref=refs%%2Fheads%%2Fmaster" %
                urlutils.local_path_to_url(os.path.abspath(".")),
            gd.get_branch_reference())

    def test_open_existing(self):
        GitRepo.init(".")

        gd = bzrdir.BzrDir.open('.')
        self.assertIsInstance(gd, dir.LocalGitDir)

    def test_open_workingtree(self):
        GitRepo.init(".")

        gd = bzrdir.BzrDir.open('.')
        raise TestSkipped
        wt = gd.open_workingtree()
        self.assertIsInstance(wt, workingtree.GitWorkingTree)

    def test_open_workingtree_bare(self):
        GitRepo.init_bare(".")

        gd = bzrdir.BzrDir.open('.')
        self.assertRaises(errors.NoWorkingTree, gd.open_workingtree)


class TestGitDirFormat(tests.TestCase):

    def setUp(self):
        super(TestGitDirFormat, self).setUp()
        self.format = dir.LocalGitControlDirFormat()

    def test_get_format_description(self):
        self.assertEquals("Local Git Repository",
                          self.format.get_format_description())

    def test_eq(self):
        format2 = dir.LocalGitControlDirFormat()
        self.assertEquals(self.format, format2)
        self.assertEquals(self.format, self.format)
        bzr_format = bzrdir.format_registry.make_bzrdir("default")
        self.assertNotEquals(self.format, bzr_format)




# Copyright (C) 2010 Jelmer Vernooij <jelmer@samba.org>
# -*- encoding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""Tests for roundtripping text parsing."""


from ....tests import TestCase

from ..roundtrip import (
    CommitSupplement,
    deserialize_fileid_map,
    extract_bzr_metadata,
    generate_roundtripping_metadata,
    inject_bzr_metadata,
    parse_roundtripping_metadata,
    serialize_fileid_map,
    )


class RoundtripTests(TestCase):

    def test_revid(self):
        md = parse_roundtripping_metadata("revision-id: foo\n")
        self.assertEquals("foo", md.revision_id)

    def test_parent_ids(self):
        md = parse_roundtripping_metadata("parent-ids: foo bar\n")
        self.assertEquals(("foo", "bar"), md.explicit_parent_ids)

    def test_properties(self):
        md = parse_roundtripping_metadata("property-foop: blar\n")
        self.assertEquals({"foop": "blar"}, md.properties)


class FormatTests(TestCase):

    def test_revid(self):
        metadata = CommitSupplement()
        metadata.revision_id = "bla"
        self.assertEquals("revision-id: bla\n",
            generate_roundtripping_metadata(metadata, "utf-8"))

    def test_parent_ids(self):
        metadata = CommitSupplement()
        metadata.explicit_parent_ids = ("foo", "bar")
        self.assertEquals("parent-ids: foo bar\n",
            generate_roundtripping_metadata(metadata, "utf-8"))

    def test_properties(self):
        metadata = CommitSupplement()
        metadata.properties = {"foo": "bar"}
        self.assertEquals("property-foo: bar\n",
            generate_roundtripping_metadata(metadata, "utf-8"))

    def test_empty(self):
        metadata = CommitSupplement()
        self.assertEquals("",
            generate_roundtripping_metadata(metadata, "utf-8"))


class ExtractMetadataTests(TestCase):

    def test_roundtrip(self):
        (msg, metadata) = extract_bzr_metadata("""Foo
--BZR--
revision-id: foo
""")
        self.assertEquals("Foo", msg)
        self.assertEquals("foo", metadata.revision_id)


class GenerateMetadataTests(TestCase):

    def test_roundtrip(self):
        metadata = CommitSupplement()
        metadata.revision_id = "myrevid"
        msg = inject_bzr_metadata("Foo", metadata, "utf-8")
        self.assertEquals("""Foo
--BZR--
revision-id: myrevid
""", msg)

    def test_no_metadata(self):
        metadata = CommitSupplement()
        msg = inject_bzr_metadata("Foo", metadata, "utf-8")
        self.assertEquals("Foo", msg)


class FileIdRoundTripTests(TestCase):

    def test_deserialize(self):
        self.assertEquals({"bar/bla": "fid"},
            deserialize_fileid_map("bar/bla\0fid\n"))

    def test_serialize(self):
        self.assertEquals(["bar/bla\0fid\n"],
            serialize_fileid_map({"bar/bla": "fid"}))




# Copyright (C) 2006-2009 Canonical Ltd

# Authors: Robert Collins <robert.collins@canonical.com>
#          Jelmer Vernooij <jelmer@samba.org>
#          John Carr <john.carr@unrouted.co.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""A GIT branch and repository format implementation for bzr."""

from __future__ import absolute_import

import os
import sys

import bzrlib
import bzrlib.api

from .info import (
    bzr_compatible_versions,
    bzr_plugin_version as version_info,
    dulwich_minimum_version,
    )

if version_info[3] == 'final':
    version_string = '%d.%d.%d' % version_info[:3]
else:
    version_string = '%d.%d.%d%s%d' % version_info
__version__ = version_string

bzrlib.api.require_any_api(bzrlib, bzr_compatible_versions)

try:
    from ...i18n import load_plugin_translations
except ImportError: # No translations for bzr < 2.5
    gettext = lambda x: x
else:
    translation = load_plugin_translations("bzr-git")
    gettext = translation.gettext

from ... import (
    errors as bzr_errors,
    trace,
    )

from ...controldir import (
    ControlDirFormat,
    Prober,
    format_registry,
    network_format_registry as controldir_network_format_registry,
    )

from ...transport import (
    register_lazy_transport,
    register_transport_proto,
    transport_server_registry,
    )
from ...commands import (
    plugin_cmds,
    )


if getattr(sys, "frozen", None):
    # allow import additional libs from ./_lib for bzr.exe only
    sys.path.append(os.path.normpath(
        os.path.join(os.path.dirname(__file__), '_lib')))


def import_dulwich():
    try:
        from dulwich import __version__ as dulwich_version
    except ImportError:
        raise bzr_errors.DependencyNotPresent("dulwich",
            "bzr-git: Please install dulwich, https://launchpad.net/dulwich")
    else:
        if dulwich_version < dulwich_minimum_version:
            raise bzr_errors.DependencyNotPresent("dulwich",
                "bzr-git: Dulwich is too old; at least %d.%d.%d is required" %
                    dulwich_minimum_version)


_versions_checked = False
def lazy_check_versions():
    global _versions_checked
    if _versions_checked:
        return
    import_dulwich()
    _versions_checked = True

format_registry.register_lazy('git',
    "bzrlib.plugins.git.dir", "LocalGitControlDirFormat",
    help='GIT repository.', native=False, experimental=False,
    )

format_registry.register_lazy('git-bare',
    "bzrlib.plugins.git.dir", "BareLocalGitControlDirFormat",
    help='Bare GIT repository (no working tree).', native=False,
    experimental=False,
    )

from ...revisionspec import (RevisionSpec_dwim, revspec_registry)
revspec_registry.register_lazy("git:", "bzrlib.plugins.git.revspec",
    "RevisionSpec_git")
RevisionSpec_dwim.append_possible_lazy_revspec(
    "bzrlib.plugins.git.revspec", "RevisionSpec_git")


class LocalGitProber(Prober):

    def probe_transport(self, transport):
        try:
            external_url = transport.external_url()
        except bzr_errors.InProcessTransport:
            raise bzr_errors.NotBranchError(path=transport.base)
        if (external_url.startswith("http:") or
            external_url.startswith("https:")):
            # Already handled by RemoteGitProber
            raise bzr_errors.NotBranchError(path=transport.base)
        from ... import urlutils
        if urlutils.split(transport.base)[1] == ".git":
            raise bzr_errors.NotBranchError(path=transport.base)
        if not transport.has_any(['objects', '.git/objects']):
            raise bzr_errors.NotBranchError(path=transport.base)
        lazy_check_versions()
        from .dir import (
            BareLocalGitControlDirFormat,
            LocalGitControlDirFormat,
            )
        if transport.has_any(['.git/objects']):
            return LocalGitControlDirFormat()
        if transport.has('info') and transport.has('objects'):
            return BareLocalGitControlDirFormat()
        raise bzr_errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        from .dir import (
            BareLocalGitControlDirFormat,
            LocalGitControlDirFormat,
            )
        return set([BareLocalGitControlDirFormat(), LocalGitControlDirFormat()])


class RemoteGitProber(Prober):

    def probe_http_transport(self, transport):
        from ... import urlutils
        base_url, _ = urlutils.split_segment_parameters(transport.external_url())
        url = urlutils.join(base_url, "info/refs") + "?service=git-upload-pack"
        from ...transport.http._urllib import HttpTransport_urllib, Request
        if isinstance(transport, HttpTransport_urllib):
            req = Request('GET', url, accepted_errors=[200, 403, 404, 405],
                          headers={"Content-Type": "application/x-git-upload-pack-request"})
            req.follow_redirections = True
            resp = transport._perform(req)
            if resp.code in (404, 405):
                raise bzr_errors.NotBranchError(transport.base)
            headers = resp.headers
            refs_text = resp.read()
        else:
            try:
                from ...transport.http._pycurl import PyCurlTransport
            except bzr_errors.DependencyNotPresent:
                raise bzr_errors.NotBranchError(transport.base)
            else:
                import pycurl
                from cStringIO import StringIO
                if isinstance(transport, PyCurlTransport):
                    conn = transport._get_curl()
                    conn.setopt(pycurl.URL, url)
                    conn.setopt(pycurl.FOLLOWLOCATION, 1)
                    transport._set_curl_options(conn)
                    conn.setopt(pycurl.HTTPGET, 1)
                    header = StringIO()
                    data = StringIO()
                    conn.setopt(pycurl.HEADERFUNCTION, header.write)
                    conn.setopt(pycurl.WRITEFUNCTION, data.write)
                    transport._curl_perform(conn, header,
                        ["Content-Type: application/x-git-upload-pack-request"])
                    code = conn.getinfo(pycurl.HTTP_CODE)
                    if code in (404, 405):
                        raise bzr_errors.NotBranchError(transport.base)
                    if code != 200:
                        raise bzr_errors.InvalidHttpResponse(transport._path,
                            str(code))
                    headers = transport._parse_headers(header)
                else:
                    raise bzr_errors.NotBranchError(transport.base)
                refs_text = data.getvalue()
        ct = headers.getheader("Content-Type")
        if ct is None:
            raise bzr_errors.NotBranchError(transport.base)
        if ct.startswith("application/x-git"):
            from .remote import RemoteGitControlDirFormat
            return RemoteGitControlDirFormat()
        else:
            from .dir import (
                BareLocalGitControlDirFormat,
                )
            ret = BareLocalGitControlDirFormat()
            ret._refs_text = refs_text
            return ret

    def probe_transport(self, transport):
        try:
            external_url = transport.external_url()
        except bzr_errors.InProcessTransport:
            raise bzr_errors.NotBranchError(path=transport.base)

        if (external_url.startswith("http:") or
            external_url.startswith("https:")):
            return self.probe_http_transport(transport)

        if (not external_url.startswith("git://") and
            not external_url.startswith("git+")):
            raise bzr_errors.NotBranchError(transport.base)

        # little ugly, but works
        from .remote import (
            GitSmartTransport,
            RemoteGitControlDirFormat,
            )
        if isinstance(transport, GitSmartTransport):
            return RemoteGitControlDirFormat()
        raise bzr_errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        from .remote import RemoteGitControlDirFormat
        return set([RemoteGitControlDirFormat()])


ControlDirFormat.register_prober(LocalGitProber)
ControlDirFormat._server_probers.append(RemoteGitProber)

register_transport_proto('git://',
        help="Access using the Git smart server protocol.")
register_transport_proto('git+ssh://',
        help="Access using the Git smart server protocol over SSH.")

register_lazy_transport("git://", __name__ + '.remote',
                        'TCPGitSmartTransport')
register_lazy_transport("git+ssh://", __name__ + '.remote',
                        'SSHGitSmartTransport')


plugin_cmds.register_lazy("cmd_git_import", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_object", ["git-objects", "git-cat"],
    __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_refs", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_apply", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_push_pristine_tar_deltas",
        ['git-push-pristine-tar', 'git-push-pristine'],
    __name__ + ".commands")

def extract_git_foreign_revid(rev):
    try:
        foreign_revid = rev.foreign_revid
    except AttributeError:
        from .mapping import mapping_registry
        foreign_revid, mapping = \
            mapping_registry.parse_revision_id(rev.revision_id)
        return foreign_revid
    else:
        from .mapping import foreign_vcs_git
        if rev.mapping.vcs == foreign_vcs_git:
            return foreign_revid
        else:
            raise bzr_errors.InvalidRevisionId(rev.revision_id, None)


def update_stanza(rev, stanza):
    mapping = getattr(rev, "mapping", None)
    try:
        git_commit = extract_git_foreign_revid(rev)
    except bzr_errors.InvalidRevisionId:
        pass
    else:
        stanza.add("git-commit", git_commit)

from ...hooks import install_lazy_named_hook
install_lazy_named_hook("bzrlib.version_info_formats.format_rio",
    "RioVersionInfoBuilder.hooks", "revision", update_stanza,
    "git commits")


transport_server_registry.register_lazy('git',
    __name__ + '.server',
    'serve_git',
    'Git Smart server protocol over TCP. (default port: 9418)')

transport_server_registry.register_lazy('git-receive-pack',
    __name__ + '.server',
    'serve_git_receive_pack',
    help='Git Smart server receive pack command. (inetd mode only)')
transport_server_registry.register_lazy('git-upload-pack',
    __name__ + 'git.server',
    'serve_git_upload_pack',
    help='Git Smart server upload pack command. (inetd mode only)')

from ...repository import (
    format_registry as repository_format_registry,
    network_format_registry as repository_network_format_registry,
    )
repository_network_format_registry.register_lazy('git',
    __name__ + '.repository', 'GitRepositoryFormat')

register_extra_lazy_repository_format = getattr(repository_format_registry,
    "register_extra_lazy")
register_extra_lazy_repository_format(__name__ + '.repository',
    'GitRepositoryFormat')

from ...branch import (
    network_format_registry as branch_network_format_registry,
    )
branch_network_format_registry.register_lazy('git',
    __name__ + '.branch', 'GitBranchFormat')

from ...branch import (
    format_registry as branch_format_registry,
    )
branch_format_registry.register_extra_lazy(
    __name__ + '.branch',
    'GitBranchFormat',
    )

from ...workingtree import (
    format_registry as workingtree_format_registry,
    )
workingtree_format_registry.register_extra_lazy(
    __name__ + '.workingtree',
    'GitWorkingTreeFormat',
    )

controldir_network_format_registry.register_lazy('git',
    __name__ + ".dir", "GitControlDirFormat")


try:
    from ...registry import register_lazy
except ImportError:
    from ...diff import format_registry as diff_format_registry
    diff_format_registry.register_lazy('git', __name__ + '.send',
        'GitDiffTree', 'Git am-style diff format')

    from ...send import (
        format_registry as send_format_registry,
        )
    send_format_registry.register_lazy('git', __name__ + '.send',
                                       'send_git', 'Git am-style diff format')

    from ...directory_service import directories
    directories.register_lazy('github:', __name__ + '.directory',
                              'GitHubDirectory',
                              'GitHub directory.')
    directories.register_lazy('git@github.com:', __name__ + '.directory',
                              'GitHubDirectory',
                              'GitHub directory.')

    from ...help_topics import (
        topic_registry,
        )
    topic_registry.register_lazy('git', __name__ + '.help', 'help_git',
        'Using Bazaar with Git')

    from ...foreign import (
        foreign_vcs_registry,
        )
    foreign_vcs_registry.register_lazy("git",
        __name__ + ".mapping", "foreign_vcs_git", "Stupid content tracker")
else:
    register_lazy("bzrlib.diff", "format_registry",
        'git', __name__ + '.send', 'GitDiffTree',
        'Git am-style diff format')
    register_lazy("bzrlib.send", "format_registry",
        'git', __name__ + '.send', 'send_git',
        'Git am-style diff format')
    register_lazy('bzrlib.directory_service', 'directories', 'github:',
            __name__ + '.directory', 'GitHubDirectory',
            'GitHub directory.')
    register_lazy('bzrlib.directory_service', 'directories',
            'git@github.com:', 'bzrlib.plugins.git.directory',
            'GitHubDirectory', 'GitHub directory.')
    register_lazy('bzrlib.help_topics', 'topic_registry',
            'git', __name__ + '.help', 'help_git',
            'Using Bazaar with Git')
    register_lazy('bzrlib.foreign', 'foreign_vcs_registry', "git",
        __name__ + ".mapping", "foreign_vcs_git", "Stupid content tracker")

def update_git_cache(repository, revid):
    """Update the git cache after a local commit."""
    if getattr(repository, "_git", None) is not None:
        return # No need to update cache for git repositories

    if not repository.control_transport.has("git"):
        return # No existing cache, don't bother updating
    try:
        lazy_check_versions()
    except bzr_errors.DependencyNotPresent, e:
        # dulwich is probably missing. silently ignore
        trace.mutter("not updating git map for %r: %s",
            repository, e)

    from .object_store import BazaarObjectStore
    store = BazaarObjectStore(repository)
    store.lock_write()
    try:
        try:
            parent_revisions = set(repository.get_parent_map([revid])[revid])
        except KeyError:
            # Isn't this a bit odd - how can a revision that was just committed be missing?
            return
        missing_revisions = store._missing_revisions(parent_revisions)
        if not missing_revisions:
            # Only update if the cache was up to date previously
            store._update_sha_map_revision(revid)
    finally:
        store.unlock()


def post_commit_update_cache(local_branch, master_branch, old_revno, old_revid,
        new_revno, new_revid):
    if local_branch is not None:
        update_git_cache(local_branch.repository, new_revid)
    update_git_cache(master_branch.repository, new_revid)


def loggerhead_git_hook(branch_app, environ):
    branch = branch_app.branch
    config_stack = branch.get_config_stack()
    if config_stack.get('http_git'):
        return None
    from .server import git_http_hook
    return git_http_hook(branch, environ['REQUEST_METHOD'],
        environ['PATH_INFO'])

install_lazy_named_hook("bzrlib.branch",
    "Branch.hooks", "post_commit", post_commit_update_cache,
    "git cache")
install_lazy_named_hook("bzrlib.plugins.loggerhead.apps.branch",
    "BranchWSGIApp.hooks", "controller",
    loggerhead_git_hook, "git support")


from ...config import (
    option_registry,
    Option,
    bool_from_store,
    )

option_registry.register(
    Option('git.http',
           default=None, from_unicode=bool_from_store, invalid='warning',
           help='''\
Allow fetching of Git packs over HTTP.

This enables support for fetching Git packs over HTTP in Loggerhead.
'''))

def test_suite():
    from . import tests
    return tests.test_suite()



# Copyright (C) 2009 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Tests for GitShaMap."""

from dulwich.objects import (
    Blob,
    Commit,
    Tree,
    )

import os
import stat

from .... import osutils

from ....inventory import (
    InventoryFile,
    InventoryDirectory,
    ROOT_ID,
    )

from ....revision import (
    Revision,
    )

from ....tests import (
    TestCase,
    TestCaseInTempDir,
    UnavailableFeature,
    )
from ....transport import (
    get_transport,
    )

from ..cache import (
    DictBzrGitCache,
    IndexBzrGitCache,
    IndexGitCacheFormat,
    SqliteBzrGitCache,
    TdbBzrGitCache,
    )

class TestGitShaMap:

    def _get_test_commit(self):
        c = Commit()
        c.committer = "Jelmer <jelmer@samba.org>"
        c.commit_time = 0
        c.commit_timezone = 0
        c.author = "Jelmer <jelmer@samba.org>"
        c.author_time = 0
        c.author_timezone = 0
        c.message = "Teh foo bar"
        c.tree = "cc9462f7f8263ef5adfbeff2fb936bb36b504cba"
        return c

    def test_commit(self):
        self.map.start_write_group()
        updater = self.cache.get_updater(Revision("myrevid"))
        c = self._get_test_commit()
        updater.add_object(c, {
            "testament3-sha1": "cc9462f7f8263ef5adf8eff2fb936bb36b504cba"},
            None)
        updater.finish()
        self.map.commit_write_group()
        self.assertEquals(
            [("commit", ("myrevid",
                "cc9462f7f8263ef5adfbeff2fb936bb36b504cba",
                {"testament3-sha1": "cc9462f7f8263ef5adf8eff2fb936bb36b504cba"},
                ))],
            list(self.map.lookup_git_sha(c.id)))
        self.assertEquals(c.id, self.map.lookup_commit("myrevid"))

    def test_lookup_notfound(self):
        self.assertRaises(KeyError, list,
            self.map.lookup_git_sha("5686645d49063c73d35436192dfc9a160c672301"))

    def test_blob(self):
        self.map.start_write_group()
        updater = self.cache.get_updater(Revision("myrevid"))
        updater.add_object(self._get_test_commit(), { "testament3-sha1": "Test" }, None)
        b = Blob()
        b.data = "TEH BLOB"
        updater.add_object(b, ("myfileid", "myrevid"), None)
        updater.finish()
        self.map.commit_write_group()
        self.assertEquals(
            [("blob", ("myfileid", "myrevid"))],
            list(self.map.lookup_git_sha(b.id)))
        self.assertEquals(b.id,
            self.map.lookup_blob_id("myfileid", "myrevid"))

    def test_tree(self):
        self.map.start_write_group()
        updater = self.cache.get_updater(Revision("myrevid"))
        updater.add_object(self._get_test_commit(), {
            "testament3-sha1": "mytestamentsha" }, None)
        t = Tree()
        t.add("somename", stat.S_IFREG, Blob().id)
        updater.add_object(t, ("fileid", ), "")
        updater.finish()
        self.map.commit_write_group()
        self.assertEquals([("tree", ("fileid", "myrevid"))],
            list(self.map.lookup_git_sha(t.id)))
        # It's possible for a backend to not implement lookup_tree
        try:
            self.assertEquals(t.id,
                self.map.lookup_tree_id("fileid", "myrevid"))
        except NotImplementedError:
            pass

    def test_revids(self):
        self.map.start_write_group()
        updater = self.cache.get_updater(Revision("myrevid"))
        c = self._get_test_commit()
        updater.add_object(c, {"testament3-sha1": "mtestament"}, None)
        updater.finish()
        self.map.commit_write_group()
        self.assertEquals(["myrevid"], list(self.map.revids()))

    def test_missing_revisions(self):
        self.map.start_write_group()
        updater = self.cache.get_updater(Revision("myrevid"))
        c = self._get_test_commit()
        updater.add_object(c, {"testament3-sha1": "testament"}, None)
        updater.finish()
        self.map.commit_write_group()
        self.assertEquals(set(["lala", "bla"]),
            set(self.map.missing_revisions(["myrevid", "lala", "bla"])))


class DictGitShaMapTests(TestCase,TestGitShaMap):

    def setUp(self):
        TestCase.setUp(self)
        self.cache = DictBzrGitCache()
        self.map = self.cache.idmap


class SqliteGitShaMapTests(TestCaseInTempDir,TestGitShaMap):

    def setUp(self):
        TestCaseInTempDir.setUp(self)
        self.cache = SqliteBzrGitCache(os.path.join(self.test_dir, 'foo.db'))
        self.map = self.cache.idmap


class TdbGitShaMapTests(TestCaseInTempDir,TestGitShaMap):

    def setUp(self):
        TestCaseInTempDir.setUp(self)
        try:
            self.cache = TdbBzrGitCache(
                os.path.join(self.test_dir, 'foo.tdb').encode(osutils._fs_enc))
        except ImportError:
            raise UnavailableFeature("Missing tdb")
        self.map = self.cache.idmap


class IndexGitShaMapTests(TestCaseInTempDir,TestGitShaMap):

    def setUp(self):
        TestCaseInTempDir.setUp(self)
        transport = get_transport(self.test_dir)
        IndexGitCacheFormat().initialize(transport)
        self.cache = IndexBzrGitCache(transport)
        self.map = self.cache.idmap



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""A grouping of Exceptions for bzr-git"""

from __future__ import absolute_import

from dulwich import errors as git_errors

from ... import errors as bzr_errors


class BzrGitError(bzr_errors.BzrError):
    """The base-level exception for bzr-git errors."""


class NoSuchRef(BzrGitError):
    """Raised when a ref can not be found."""

    _fmt = "The ref %(ref)s was not found in the repository at %(location)s."

    def __init__(self, ref, location, present_refs=None):
        self.ref = ref
        self.location = location
        self.present_refs = present_refs


def convert_dulwich_error(error):
    """Convert a Dulwich error to a Bazaar error."""

    if isinstance(error, git_errors.HangupException):
        raise bzr_errors.ConnectionReset(error.msg, "")
    raise error


class NoPushSupport(bzr_errors.BzrError):
    _fmt = "Push is not yet supported for bzr-git. Try dpush instead."


class GitSmartRemoteNotSupported(bzr_errors.UnsupportedOperation):
    _fmt = "This operation is not supported by the Git smart server protocol."


class UnknownCommitExtra(bzr_errors.BzrError):
    _fmt = "Unknown extra fields in %(object)r: %(fields)r."

    def __init__(self, object, fields):
        bzr_errors.BzrError.__init__(self)
        self.object = object
        self.fields = ",".join(fields)


class UnknownMercurialCommitExtra(bzr_errors.BzrError):
    _fmt = "Unknown mercurial extra fields in %(object)r: %(fields)r."

    def __init__(self, object, fields):
        bzr_errors.BzrError.__init__(self)
        self.object = object
        self.fields = ",".join(fields)



# Copyright (C) 2010-2012 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""A Git repository implementation that uses a Bazaar transport."""

from __future__ import absolute_import

from cStringIO import StringIO

import os
import sys
import urllib

from dulwich.errors import (
    NotGitRepository,
    NoIndexPresent,
    )
from dulwich.objects import (
    ShaFile,
    )
from dulwich.object_store import (
    PackBasedObjectStore,
    PACKDIR,
    )
from dulwich.pack import (
    MemoryPackIndex,
    PackData,
    Pack,
    iter_sha1,
    load_pack_index_file,
    write_pack_data,
    write_pack_index_v2,
    )
from dulwich.repo import (
    BaseRepo,
    RefsContainer,
    BASE_DIRECTORIES,
    INDEX_FILENAME,
    OBJECTDIR,
    REFSDIR,
    SYMREF,
    check_ref_format,
    read_packed_refs_with_peeled,
    read_packed_refs,
    write_packed_refs,
    )

from ... import (
    transport as _mod_transport,
    )
from ...errors import (
    FileExists,
    NoSuchFile,
    TransportNotPossible,
    )


class TransportRefsContainer(RefsContainer):
    """Refs container that reads refs from a transport."""

    def __init__(self, transport):
        self.transport = transport
        self._packed_refs = None
        self._peeled_refs = None

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.transport)

    def _ensure_dir_exists(self, path):
        for n in range(path.count("/")):
            dirname = "/".join(path.split("/")[:n+1])
            try:
                self.transport.mkdir(dirname)
            except FileExists:
                pass

    def subkeys(self, base):
        keys = set()
        try:
            iter_files = self.transport.clone(base).iter_files_recursive()
            keys.update(("%s/%s" % (base, urllib.unquote(refname))).strip("/") for
                    refname in iter_files if check_ref_format("%s/%s" % (base, refname)))
        except (TransportNotPossible, NoSuchFile):
            pass
        for key in self.get_packed_refs():
            if key.startswith(base):
                keys.add(key[len(base):].strip("/"))
        return keys

    def allkeys(self):
        keys = set()
        try:
            self.transport.get_bytes("HEAD")
        except NoSuchFile:
            pass
        else:
            keys.add("HEAD")
        try:
            iter_files = list(self.transport.clone("refs").iter_files_recursive())
            for filename in iter_files:
                refname = "refs/%s" % urllib.unquote(filename)
                if check_ref_format(refname):
                    keys.add(refname)
        except (TransportNotPossible, NoSuchFile):
            pass
        keys.update(self.get_packed_refs())
        return keys

    def get_packed_refs(self):
        """Get contents of the packed-refs file.

        :return: Dictionary mapping ref names to SHA1s

        :note: Will return an empty dictionary when no packed-refs file is
            present.
        """
        # TODO: invalidate the cache on repacking
        if self._packed_refs is None:
            # set both to empty because we want _peeled_refs to be
            # None if and only if _packed_refs is also None.
            self._packed_refs = {}
            self._peeled_refs = {}
            try:
                f = self.transport.get("packed-refs")
            except NoSuchFile:
                return {}
            try:
                first_line = iter(f).next().rstrip()
                if (first_line.startswith("# pack-refs") and " peeled" in
                        first_line):
                    for sha, name, peeled in read_packed_refs_with_peeled(f):
                        self._packed_refs[name] = sha
                        if peeled:
                            self._peeled_refs[name] = peeled
                else:
                    f.seek(0)
                    for sha, name in read_packed_refs(f):
                        self._packed_refs[name] = sha
            finally:
                f.close()
        return self._packed_refs

    def get_peeled(self, name):
        """Return the cached peeled value of a ref, if available.

        :param name: Name of the ref to peel
        :return: The peeled value of the ref. If the ref is known not point to a
            tag, this will be the SHA the ref refers to. If the ref may point to
            a tag, but no cached information is available, None is returned.
        """
        self.get_packed_refs()
        if self._peeled_refs is None or name not in self._packed_refs:
            # No cache: no peeled refs were read, or this ref is loose
            return None
        if name in self._peeled_refs:
            return self._peeled_refs[name]
        else:
            # Known not peelable
            return self[name]

    def read_loose_ref(self, name):
        """Read a reference file and return its contents.

        If the reference file a symbolic reference, only read the first line of
        the file. Otherwise, only read the first 40 bytes.

        :param name: the refname to read, relative to refpath
        :return: The contents of the ref file, or None if the file does not
            exist.
        :raises IOError: if any other error occurs
        """
        try:
            f = self.transport.get(name)
        except NoSuchFile:
            return None
        f = StringIO(f.read())
        try:
            header = f.read(len(SYMREF))
            if header == SYMREF:
                # Read only the first line
                return header + iter(f).next().rstrip("\r\n")
            else:
                # Read only the first 40 bytes
                return header + f.read(40-len(SYMREF))
        finally:
            f.close()

    def _remove_packed_ref(self, name):
        if self._packed_refs is None:
            return
        # reread cached refs from disk, while holding the lock

        self._packed_refs = None
        self.get_packed_refs()

        if name not in self._packed_refs:
            return

        del self._packed_refs[name]
        if name in self._peeled_refs:
            del self._peeled_refs[name]
        f = self.transport.open_write_stream("packed-refs")
        try:
            write_packed_refs(f, self._packed_refs, self._peeled_refs)
        finally:
            f.close()

    def set_symbolic_ref(self, name, other):
        """Make a ref point at another ref.

        :param name: Name of the ref to set
        :param other: Name of the ref to point at
        """
        self._check_refname(name)
        self._check_refname(other)
        self._ensure_dir_exists(name)
        self.transport.put_bytes(name, SYMREF + other + '\n')

    def set_if_equals(self, name, old_ref, new_ref):
        """Set a refname to new_ref only if it currently equals old_ref.

        This method follows all symbolic references, and can be used to perform
        an atomic compare-and-swap operation.

        :param name: The refname to set.
        :param old_ref: The old sha the refname must refer to, or None to set
            unconditionally.
        :param new_ref: The new sha the refname will refer to.
        :return: True if the set was successful, False otherwise.
        """
        try:
            realnames, _ = self.follow(name)
            realname = realnames[-1]
        except (KeyError, IndexError):
            realname = name
        self._ensure_dir_exists(realname)
        self.transport.put_bytes(realname, new_ref+"\n")
        return True

    def add_if_new(self, name, ref):
        """Add a new reference only if it does not already exist.

        This method follows symrefs, and only ensures that the last ref in the
        chain does not exist.

        :param name: The refname to set.
        :param ref: The new sha the refname will refer to.
        :return: True if the add was successful, False otherwise.
        """
        try:
            realnames, contents = self.follow(name)
            if contents is not None:
                return False
            realname = realnames[-1]
        except (KeyError, IndexError):
            realname = name
        self._check_refname(realname)
        self._ensure_dir_exists(realname)
        self.transport.put_bytes(realname, ref+"\n")
        return True

    def remove_if_equals(self, name, old_ref):
        """Remove a refname only if it currently equals old_ref.

        This method does not follow symbolic references. It can be used to
        perform an atomic compare-and-delete operation.

        :param name: The refname to delete.
        :param old_ref: The old sha the refname must refer to, or None to delete
            unconditionally.
        :return: True if the delete was successful, False otherwise.
        """
        self._check_refname(name)
        # may only be packed
        try:
            self.transport.delete(name)
        except NoSuchFile:
            pass
        self._remove_packed_ref(name)
        return True

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default


class TransportRepo(BaseRepo):

    def __init__(self, transport, bare, refs_text=None):
        self.transport = transport
        self.bare = bare
        if self.bare:
            self._controltransport = self.transport
        else:
            self._controltransport = self.transport.clone('.git')
        object_store = TransportObjectStore(
            self._controltransport.clone(OBJECTDIR))
        if refs_text is not None:
            from dulwich.repo import InfoRefsContainer # dulwich >= 0.8.2
            refs_container = InfoRefsContainer(StringIO(refs_text))
            try:
                head = TransportRefsContainer(self._controltransport).read_loose_ref("HEAD")
            except KeyError:
                pass
            else:
                refs_container._refs["HEAD"] = head
        else:
            refs_container = TransportRefsContainer(self._controltransport)
        super(TransportRepo, self).__init__(object_store,
                refs_container)

    def _determine_file_mode(self):
        # Be consistent with bzr
        if sys.platform == 'win32':
            return False
        return True

    def get_named_file(self, path):
        """Get a file from the control dir with a specific name.

        Although the filename should be interpreted as a filename relative to
        the control dir in a disk-baked Repo, the object returned need not be
        pointing to a file in that location.

        :param path: The path to the file, relative to the control dir.
        :return: An open file object, or None if the file does not exist.
        """
        try:
            return self._controltransport.get(path.lstrip('/'))
        except NoSuchFile:
            return None

    def _put_named_file(self, relpath, contents):
        self._controltransport.put_bytes(relpath, contents)

    def index_path(self):
        """Return the path to the index file."""
        return self._controltransport.local_abspath(INDEX_FILENAME)

    def open_index(self):
        """Open the index for this repository."""
        from dulwich.index import Index
        if not self.has_index():
            raise NoIndexPresent()
        return Index(self.index_path())

    def has_index(self):
        """Check if an index is present."""
        # Bare repos must never have index files; non-bare repos may have a
        # missing index file, which is treated as empty.
        return not self.bare

    def get_config(self):
        from dulwich.config import ConfigFile
        try:
            return ConfigFile.from_file(self._controltransport.get('config'))
        except NoSuchFile:
            return ConfigFile()

    def get_config_stack(self):
        from dulwich.config import StackedConfig
        backends = []
        p = self.get_config()
        if p is not None:
            backends.append(p)
            writable = p
        else:
            writable = None
        backends.extend(StackedConfig.default_backends())
        return StackedConfig(backends, writable=writable)

    def __repr__(self):
        return "<%s for %r>" % (self.__class__.__name__, self.transport)

    @classmethod
    def init(cls, transport, bare=False):
        if not bare:
            transport.mkdir(".git")
            control_transport = transport.clone(".git")
        else:
            control_transport = transport
        for d in BASE_DIRECTORIES:
            control_transport.mkdir("/".join(d))
        control_transport.mkdir(OBJECTDIR)
        TransportObjectStore.init(control_transport.clone(OBJECTDIR))
        ret = cls(transport, bare)
        ret.refs.set_symbolic_ref("HEAD", "refs/heads/master")
        ret._init_files(bare)
        return ret


class TransportObjectStore(PackBasedObjectStore):
    """Git-style object store that exists on disk."""

    def __init__(self, transport):
        """Open an object store.

        :param transport: Transport to open data from
        """
        super(TransportObjectStore, self).__init__()
        self.transport = transport
        self.pack_transport = self.transport.clone(PACKDIR)
        self._alternates = None

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.transport)

    @property
    def alternates(self):
        if self._alternates is not None:
            return self._alternates
        self._alternates = []
        for path in self._read_alternate_paths():
            # FIXME: Check path
            t = _mod_transport.get_transport_from_path(path)
            self._alternates.append(self.__class__(t))
        return self._alternates

    def _read_alternate_paths(self):
        try:
            f = self.transport.get("info/alternates")
        except NoSuchFile:
            return []
        ret = []
        try:
            for l in f.read().splitlines():
                if l[0] == "#":
                    continue
                if os.path.isabs(l):
                    continue
                ret.append(l)
            return ret
        finally:
            f.close()

    @property
    def packs(self):
        # FIXME: Never invalidates.
        if not self._pack_cache:
            self._update_pack_cache()
        return self._pack_cache.values()

    def _update_pack_cache(self):
        for pack in self._load_packs():
            self._pack_cache[pack._basename] = pack

    def _pack_names(self):
        try:
            f = self.transport.get('info/packs')
        except NoSuchFile:
            return self.pack_transport.list_dir(".")
        else:
            ret = []
            for line in f.read().splitlines():
                if not line:
                    continue
                (kind, name) = line.split(" ", 1)
                if kind != "P":
                    continue
                ret.append(name)
            return ret

    def _load_packs(self):
        ret = []
        for name in self._pack_names():
            if name.startswith("pack-") and name.endswith(".pack"):
                try:
                    size = self.pack_transport.stat(name).st_size
                except TransportNotPossible:
                    # FIXME: This reads the whole pack file at once
                    f = self.pack_transport.get(name)
                    contents = f.read()
                    pd = PackData(name, StringIO(contents), size=len(contents))
                else:
                    pd = PackData(name, self.pack_transport.get(name),
                            size=size)
                idxname = name.replace(".pack", ".idx")
                idx = load_pack_index_file(idxname, self.pack_transport.get(idxname))
                pack = Pack.from_objects(pd, idx)
                pack._basename = idxname[:-4]
                ret.append(pack)
        return ret

    def _iter_loose_objects(self):
        for base in self.transport.list_dir('.'):
            if len(base) != 2:
                continue
            for rest in self.transport.list_dir(base):
                yield base+rest

    def _split_loose_object(self, sha):
        return (sha[:2], sha[2:])

    def _remove_loose_object(self, sha):
        path = '%s/%s' % self._split_loose_object(sha)
        self.transport.delete(path)

    def _remove_pack(self, pack):
        self.pack_transport.delete(pack.data.filename)
        self.pack_transport.delete(pack.index.filename)

    def _get_loose_object(self, sha):
        path = '%s/%s' % self._split_loose_object(sha)
        try:
            return ShaFile.from_file(self.transport.get(path))
        except NoSuchFile:
            return None

    def add_object(self, obj):
        """Add a single object to this object store.

        :param obj: Object to add
        """
        (dir, file) = self._split_loose_object(obj.id)
        try:
            self.transport.mkdir(dir)
        except FileExists:
            pass
        path = "%s/%s" % (dir, file)
        if self.transport.has(path):
            return # Already there, no need to write again
        self.transport.put_bytes(path, obj.as_legacy_object())

    def move_in_pack(self, f):
        """Move a specific file containing a pack into the pack directory.

        :note: The file should be on the same file system as the
            packs directory.

        :param path: Path to the pack file.
        """
        f.seek(0)
        p = PackData("", f, len(f.getvalue()))
        entries = p.sorted_entries()
        basename = "pack-%s" % iter_sha1(entry[0] for entry in entries)
        f.seek(0)
        self.pack_transport.put_file(basename + ".pack", f)
        p._filename = basename + ".pack"
        idxfile = self.pack_transport.open_write_stream(basename + ".idx")
        try:
            write_pack_index_v2(idxfile, entries, p.get_stored_checksum())
        finally:
            idxfile.close()
        idxfile = self.pack_transport.get(basename + ".idx")
        idx = load_pack_index_file(basename+".idx", idxfile)
        final_pack = Pack.from_objects(p, idx)
        final_pack._basename = basename
        self._add_known_pack(basename, final_pack)
        return final_pack

    def add_thin_pack(self):
        """Add a new thin pack to this object store.

        Thin packs are packs that contain deltas with parents that exist
        in a different pack.
        """
        from cStringIO import StringIO
        f = StringIO()
        def commit():
            if len(f.getvalue()) > 0:
                return self.move_in_thin_pack(f)
            else:
                return None
        return f, commit

    def move_in_thin_pack(self, f):
        """Move a specific file containing a pack into the pack directory.

        :note: The file should be on the same file system as the
            packs directory.

        :param path: Path to the pack file.
        """
        f.seek(0)
        data = PackData.from_file(self.get_raw, f, len(f.getvalue()))
        idx = MemoryPackIndex(data.sorted_entries(), data.get_stored_checksum())
        p = Pack.from_objects(data, idx)

        pack_sha = idx.objects_sha1()

        datafile = self.pack_transport.open_write_stream(
                "pack-%s.pack" % pack_sha)
        try:
            entries, data_sum = write_pack_data(datafile, p.pack_tuples())
        finally:
            datafile.close()
        entries.sort()
        idxfile = self.pack_transport.open_write_stream(
            "pack-%s.idx" % pack_sha)
        try:
            write_pack_index_v2(idxfile, data.sorted_entries(), data_sum)
        finally:
            idxfile.close()
        basename = "pack-%s" % pack_sha
        final_pack = Pack(basename)
        self._add_known_pack(basename, final_pack)
        return final_pack

    def add_pack(self):
        """Add a new pack to this object store.

        :return: Fileobject to write to and a commit function to
            call when the pack is finished.
        """
        from cStringIO import StringIO
        f = StringIO()
        def commit():
            if len(f.getvalue()) > 0:
                return self.move_in_pack(f)
            else:
                return None
        def abort():
            return None
        return f, commit, abort

    @classmethod
    def init(cls, transport):
        transport.mkdir('info')
        transport.mkdir(PACKDIR)
        return cls(transport)

    def _remove_pack(self, pack):
        self.pack_transport.delete(pack.data.filename)
        self.pack_transport.delete(os.path.basename(pack.index.path))



# Copyright (C) 2010 Jelmer Vernooij
# vim: encoding=utf-8
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""Tests for ref handling."""

from .... import tests

from ...git import refs


class BranchNameRefConversionTests(tests.TestCase):

    def test_head(self):
        self.assertEquals("", refs.ref_to_branch_name("HEAD"))
        self.assertEquals("HEAD", refs.branch_name_to_ref(""))

    def test_tag(self):
        self.assertRaises(ValueError, refs.ref_to_branch_name, "refs/tags/FOO")

    def test_branch(self):
        self.assertEquals("frost", refs.ref_to_branch_name("refs/heads/frost"))
        self.assertEquals("refs/heads/frost", refs.branch_name_to_ref("frost"))



#!/usr/bin/env python
# vim: expandtab

# Copyright (C) 2011 Jelmer Vernooij <jelmer@apache.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from cStringIO import StringIO
import os

from dulwich.repo import Repo

from ....tests import (
    TestCaseWithTransport,
    TestSkipped,
    )

from ..object_store import get_object_store
from ..git_remote_helper import (
    RemoteHelper,
    open_local_dir,
    fastexporter,
    fetch,
    )


def map_to_git_sha1(dir, bzr_revid):
    object_store = get_object_store(dir.open_repository())
    object_store.lock_read()
    try:
        return object_store._lookup_revision_sha1(bzr_revid)
    finally:
        object_store.unlock()


class OpenLocalDirTests(TestCaseWithTransport):

    def test_from_env(self):
        self.make_branch_and_tree('bla', format='git')
        self.overrideEnv('GIT_DIR', os.path.join(self.test_dir, 'bla'))
        open_local_dir()

    def test_from_env_dir(self):
        self.make_branch_and_tree('bla', format='git')
        self.overrideEnv('GIT_DIR', os.path.join(self.test_dir, 'bla', '.git'))
        open_local_dir()

    def test_from_dir(self):
        self.make_branch_and_tree('.', format='git')
        open_local_dir()


class FetchTests(TestCaseWithTransport):

    def setUp(self):
        super(FetchTests, self).setUp()
        self.local_dir = self.make_branch_and_tree('local', format='git').bzrdir
        self.remote_tree = self.make_branch_and_tree('remote')
        self.remote_dir = self.remote_tree.bzrdir
        self.shortname = 'bzr'

    def fetch(self, wants):
        outf = StringIO()
        fetch(outf, wants, self.shortname, self.remote_dir, self.local_dir)
        return outf.getvalue()

    def test_no_wants(self):
        r = self.fetch([])
        self.assertEquals("\n", r)

    def test_simple(self):
        self.build_tree(['remote/foo'])
        self.remote_tree.add("foo")
        revid = self.remote_tree.commit("msg")
        git_sha1 = map_to_git_sha1(self.remote_dir, revid)
        out = self.fetch([(git_sha1, 'HEAD')])
        self.assertEquals(out, "\n")
        r = Repo('local')
        self.assertTrue(git_sha1 in r.object_store)
        self.assertEquals({'HEAD': '0000000000000000000000000000000000000000'}, r.get_refs())


class RemoteHelperTests(TestCaseWithTransport):

    def setUp(self):
        super(RemoteHelperTests, self).setUp()
        self.local_dir = self.make_branch_and_tree('local', format='git').bzrdir
        self.remote_tree = self.make_branch_and_tree('remote')
        self.remote_dir = self.remote_tree.bzrdir
        self.shortname = 'bzr'
        self.helper = RemoteHelper(self.local_dir, self.shortname, self.remote_dir)

    def test_capabilities(self):
        f = StringIO()
        self.helper.cmd_capabilities(f, [])
        capabs = f.getvalue()
        base = "fetch\noption\npush\n"
        self.assertTrue(capabs in (base+"\n", base+"import\n\n"), capabs)

    def test_option(self):
        f = StringIO()
        self.helper.cmd_option(f, [])
        self.assertEquals("unsupported\n", f.getvalue())

    def test_list_basic(self):
        f = StringIO()
        self.helper.cmd_list(f, [])
        self.assertEquals(
            '0000000000000000000000000000000000000000 HEAD\n\n',
            f.getvalue())

    def test_import(self):
        if fastexporter is None:
            raise TestSkipped("bzr-fastimport not available")
        self.build_tree_contents([("remote/afile", "somecontent")])
        self.remote_tree.add(["afile"])
        self.remote_tree.commit("A commit message", timestamp=1330445983,
            timezone=0, committer='Somebody <jrandom@example.com>')
        f = StringIO()
        self.helper.cmd_import(f, ["import", "refs/heads/master"])
        self.assertEquals(
            'commit refs/heads/master\n'
            'mark :1\n'
            'committer Somebody <jrandom@example.com> 1330445983 +0000\n'
            'data 16\n'
            'A commit message\n'
            'M 644 inline afile\n'
            'data 11\n'
            'somecontent\n',
            f.getvalue())



# Copyright (C) 2010 Jelmer Vernooij
# Copyright (C) 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Tests for Git working trees."""

from .... import (
    conflicts as _mod_conflicts,
    )
from ....tests import TestCaseWithTransport


class GitWorkingTreeTests(TestCaseWithTransport):

    def setUp(self):
        super(GitWorkingTreeTests, self).setUp()
        self.tree = self.make_branch_and_tree('.', format="git")

    def test_conflicts(self):
        self.assertIsInstance(self.tree.conflicts(),
            _mod_conflicts.ConflictList)



# Copyright (C) 2009 Jelmer Vernooij <jelmer@samba.org>
# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from dulwich.objects import (
    Blob,
    Tag,
    Tree,
    S_IFGITLINK,
    )
from dulwich.repo import (
    Repo as GitRepo,
    )
import os
import stat
import time

from .... import (
    knit,
    osutils,
    versionedfile,
    )
from ....branch import (
    Branch,
    )
from ....bzrdir import (
    BzrDir,
    )
from ....inventory import (
    Inventory,
    )
from ....repository import (
    Repository,
    )
from ....tests import (
    TestCaseWithTransport,
    )

from ..fetch import (
    import_git_blob,
    import_git_tree,
    import_git_submodule,
    )
from ..mapping import (
    BzrGitMappingv1,
    DEFAULT_FILE_MODE,
    )
from . import (
    GitBranchBuilder,
    )


class RepositoryFetchTests(object):

    def make_git_repo(self, path):
        os.mkdir(path)
        return GitRepo.init(os.path.abspath(path))

    def clone_git_repo(self, from_url, to_url, revision_id=None):
        oldrepos = self.open_git_repo(from_url)
        dir = BzrDir.create(to_url)
        newrepos = dir.create_repository()
        oldrepos.copy_content_into(newrepos, revision_id=revision_id)
        return newrepos

    def test_empty(self):
        self.make_git_repo("d")
        newrepos = self.clone_git_repo("d", "f")
        self.assertEquals([], newrepos.all_revision_ids())

    def make_onerev_branch(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", "foo\nbar\n", False)
        mark = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
        gitsha = bb.finish()[mark]
        os.chdir("..")
        return "d", gitsha

    def test_single_rev(self):
        path, gitsha = self.make_onerev_branch()
        oldrepo = self.open_git_repo(path)
        newrepo = self.clone_git_repo(path, "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        self.assertEquals([revid], newrepo.all_revision_ids())

    def test_single_rev_specific(self):
        path, gitsha = self.make_onerev_branch()
        oldrepo = self.open_git_repo(path)
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        newrepo = self.clone_git_repo(path, "f", revision_id=revid)
        self.assertEquals([revid], newrepo.all_revision_ids())

    def test_incremental(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", "foo\nbar\n", False)
        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
        bb.set_file("foobar", "fooll\nbar\n", False)
        mark2 = bb.commit("Somebody <somebody@someorg.org>", "nextmsg")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        newrepo = self.clone_git_repo("d", "f", revision_id=revid1)
        self.assertEquals([revid1], newrepo.all_revision_ids())
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        newrepo.fetch(oldrepo, revision_id=revid2)
        self.assertEquals(set([revid1, revid2]), set(newrepo.all_revision_ids()))

    def test_dir_becomes_symlink(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("mylink/somefile", "foo\nbar\n", False)
        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg1")
        bb.set_symlink("mylink", "target/")
        mark2 = bb.commit("Somebody <somebody@someorg.org>", "mymsg2")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree1 = newrepo.revision_tree(revid1)
        tree2 = newrepo.revision_tree(revid2)
        fileid = tree1.path2id("mylink")
        self.assertEquals(revid1, tree1.get_file_revision(fileid))
        self.assertEquals("directory", tree1.kind(fileid))
        self.assertEquals(None, tree1.get_symlink_target(fileid))
        self.assertEquals(revid2, tree2.get_file_revision(fileid))
        self.assertEquals("symlink", tree2.kind(fileid))
        self.assertEquals("target/", tree2.get_symlink_target(fileid))

    def test_symlink_becomes_dir(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_symlink("mylink", "target/")
        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg1")
        bb.set_file("mylink/somefile", "foo\nbar\n", False)
        mark2 = bb.commit("Somebody <somebody@someorg.org>", "mymsg2")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree1 = newrepo.revision_tree(revid1)
        tree2 = newrepo.revision_tree(revid2)
        fileid = tree1.path2id("mylink")
        self.assertEquals(revid1, tree1.get_file_revision(fileid))
        self.assertEquals("symlink", tree1.kind(fileid))
        self.assertEquals("target/", tree1.get_symlink_target(fileid))
        self.assertEquals(revid2, tree2.get_file_revision(fileid))
        self.assertEquals("directory", tree2.kind(fileid))
        self.assertEquals(None, tree2.get_symlink_target(fileid))

    def test_changing_symlink(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_symlink("mylink", "target")
        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg1")
        bb.set_symlink("mylink", "target/")
        mark2 = bb.commit("Somebody <somebody@someorg.org>", "mymsg2")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree1 = newrepo.revision_tree(revid1)
        tree2 = newrepo.revision_tree(revid2)
        fileid = tree1.path2id("mylink")
        self.assertEquals(revid1, tree1.get_file_revision(fileid))
        self.assertEquals("target", tree1.get_symlink_target(fileid))
        self.assertEquals(revid2, tree2.get_file_revision(fileid))
        self.assertEquals("target/", tree2.get_symlink_target(fileid))

    def test_executable(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", "foo\nbar\n", True)
        bb.set_file("notexec", "foo\nbar\n", False)
        mark = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
        gitsha = bb.finish()[mark]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        tree = newrepo.revision_tree(revid)
        self.assertTrue(tree.has_filename("foobar"))
        self.assertEquals(True, tree.is_executable(tree.path2id("foobar")))
        self.assertTrue(tree.has_filename("notexec"))
        self.assertEquals(False, tree.is_executable(tree.path2id("notexec")))

    def test_becomes_executable(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", "foo\nbar\n", False)
        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
        bb.set_file("foobar", "foo\nbar\n", True)
        mark2 = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
        gitsha2 = bb.finish()[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree = newrepo.revision_tree(revid)
        self.assertTrue(tree.has_filename("foobar"))
        fileid = tree.path2id("foobar")
        self.assertEquals(True, tree.is_executable(fileid))
        self.assertEquals(revid, tree.get_file_revision(fileid))

    def test_into_stacked_on(self):
        r = self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file(u"foobar", "foo\n", False)
        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg1")
        gitsha1 = bb.finish()[mark1]
        os.chdir("..")
        stacked_on = self.clone_git_repo("d", "stacked-on")
        oldrepo = Repository.open("d")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        self.assertEquals([revid1], stacked_on.all_revision_ids())
        b = stacked_on.bzrdir.create_branch()
        b.generate_revision_history(revid1)
        self.assertEquals(b.last_revision(), revid1)
        tree = self.make_branch_and_tree("stacked")
        tree.branch.set_stacked_on_url(b.user_url)
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file(u"barbar", "bar\n", False)
        bb.set_file(u"foo/blie/bla", "bla\n", False)
        mark2 = bb.commit("Somebody <somebody@someorg.org>", "mymsg2")
        gitsha2 = bb.finish()[mark2]
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        os.chdir("..")
        tree.branch.fetch(Branch.open("d"))
        tree.branch.repository.check()
        self.addCleanup(tree.lock_read().unlock)
        self.assertEquals(
            set([(revid2,)]),
            tree.branch.repository.revisions.without_fallbacks().keys())
        self.assertEquals(
            set([revid1, revid2]),
            set(tree.branch.repository.all_revision_ids()))

    def test_non_ascii_characters(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file(u"foőbar", "foo\nbar\n", False)
        mark = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
        gitsha = bb.finish()[mark]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        tree = newrepo.revision_tree(revid)
        self.assertTrue(tree.has_filename(u"foőbar"))

    def test_tagged_tree(self):
        r = self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", "fooll\nbar\n", False)
        mark = bb.commit("Somebody <somebody@someorg.org>", "nextmsg")
        marks = bb.finish()
        gitsha = marks[mark]
        tag = Tag()
        tag.name = "sometag"
        tag.tag_time = int(time.time())
        tag.tag_timezone = 0
        tag.tagger = "Somebody <somebody@example.com>"
        tag.message = "Created tag pointed at tree"
        tag.object = (Tree, r[gitsha].tree)
        r.object_store.add_object(tag)
        r["refs/tags/sometag"] = tag
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        newrepo = self.clone_git_repo("d", "f")
        self.assertEquals(set([revid]), set(newrepo.all_revision_ids()))


class LocalRepositoryFetchTests(RepositoryFetchTests, TestCaseWithTransport):

    def open_git_repo(self, path):
        return Repository.open(path)


class DummyStoreUpdater(object):

    def add_object(self, obj, ie, path):
        pass

    def finish(self):
        pass


class ImportObjects(TestCaseWithTransport):

    def setUp(self):
        super(ImportObjects, self).setUp()
        self._mapping = BzrGitMappingv1()
        factory = knit.make_file_factory(True, versionedfile.PrefixMapper())
        self._texts = factory(self.get_transport('texts'))

    def test_import_blob_simple(self):
        blob = Blob.from_string("bar")
        base_inv = Inventory()
        objs = { "blobname": blob}
        ret = import_git_blob(self._texts, self._mapping, "bla", "bla",
            (None, "blobname"), 
            base_inv, None, "somerevid", [], objs.__getitem__, 
            (None, DEFAULT_FILE_MODE), DummyStoreUpdater(),
            self._mapping.generate_file_id)
        self.assertEquals(set([('bla', 'somerevid')]), self._texts.keys())
        self.assertEquals(self._texts.get_record_stream([('bla', 'somerevid')],
            "unordered", True).next().get_bytes_as("fulltext"), "bar")
        self.assertEquals(1, len(ret)) 
        self.assertEquals(None, ret[0][0])
        self.assertEquals("bla", ret[0][1])
        ie = ret[0][3]
        self.assertEquals(False, ie.executable)
        self.assertEquals("file", ie.kind)
        self.assertEquals("somerevid", ie.revision)
        self.assertEquals(osutils.sha_strings(["bar"]), ie.text_sha1)

    def test_import_tree_empty_root(self):
        base_inv = Inventory(root_id=None)
        tree = Tree()
        ret, child_modes = import_git_tree(self._texts, self._mapping, "", "",
               (None, tree.id), base_inv, 
               None, "somerevid", [], {tree.id: tree}.__getitem__,
               (None, stat.S_IFDIR), DummyStoreUpdater(),
               self._mapping.generate_file_id)
        self.assertEquals(child_modes, {})
        self.assertEquals(set([("TREE_ROOT", 'somerevid')]), self._texts.keys())
        self.assertEquals(1, len(ret))
        self.assertEquals(None, ret[0][0])
        self.assertEquals("", ret[0][1])
        ie = ret[0][3]
        self.assertEquals(False, ie.executable)
        self.assertEquals("directory", ie.kind)
        self.assertEquals({}, ie.children)
        self.assertEquals("somerevid", ie.revision)
        self.assertEquals(None, ie.text_sha1)

    def test_import_tree_empty(self):
        base_inv = Inventory()
        tree = Tree()
        ret, child_modes = import_git_tree(self._texts, self._mapping, "bla", "bla",
           (None, tree.id), base_inv, None, "somerevid", [], 
           { tree.id: tree }.__getitem__,
           (None, stat.S_IFDIR), DummyStoreUpdater(),
           self._mapping.generate_file_id)
        self.assertEquals(child_modes, {})
        self.assertEquals(set([("bla", 'somerevid')]), self._texts.keys())
        self.assertEquals(1, len(ret))
        self.assertEquals(None, ret[0][0])
        self.assertEquals("bla", ret[0][1])
        ie = ret[0][3]
        self.assertEquals("directory", ie.kind)
        self.assertEquals(False, ie.executable)
        self.assertEquals({}, ie.children)
        self.assertEquals("somerevid", ie.revision)
        self.assertEquals(None, ie.text_sha1)

    def test_import_tree_with_file(self):
        base_inv = Inventory()
        blob = Blob.from_string("bar1")
        tree = Tree()
        tree.add("foo", stat.S_IFREG | 0644, blob.id)
        objects = { blob.id: blob, tree.id: tree }
        ret, child_modes = import_git_tree(self._texts, self._mapping, "bla", "bla",
                (None, tree.id), base_inv, None, "somerevid", [],
            objects.__getitem__, (None, stat.S_IFDIR), DummyStoreUpdater(),
            self._mapping.generate_file_id)
        self.assertEquals(child_modes, {})
        self.assertEquals(2, len(ret))
        self.assertEquals(None, ret[0][0])
        self.assertEquals("bla", ret[0][1])
        self.assertEquals(None, ret[1][0])
        self.assertEquals("bla/foo", ret[1][1])
        ie = ret[0][3]
        self.assertEquals("directory", ie.kind)
        ie = ret[1][3]
        self.assertEquals("file", ie.kind)
        self.assertEquals("bla/foo", ie.file_id)
        self.assertEquals("somerevid", ie.revision)
        self.assertEquals(osutils.sha_strings(["bar1"]), ie.text_sha1)
        self.assertEquals(False, ie.executable)

    def test_import_tree_with_unusual_mode_file(self):
        base_inv = Inventory()
        blob = Blob.from_string("bar1")
        tree = Tree()
        tree.add("foo", stat.S_IFREG | 0664, blob.id)
        objects = { blob.id: blob, tree.id: tree }
        ret, child_modes = import_git_tree(self._texts, self._mapping,
            "bla", "bla", (None, tree.id), base_inv, None, "somerevid", [],
            objects.__getitem__, (None, stat.S_IFDIR), DummyStoreUpdater(),
            self._mapping.generate_file_id)
        self.assertEquals(child_modes, { "bla/foo": stat.S_IFREG | 0664 })

    def test_import_tree_with_file_exe(self):
        base_inv = Inventory(root_id=None)
        blob = Blob.from_string("bar")
        tree = Tree()
        tree.add("foo", 0100755, blob.id)
        objects = { blob.id: blob, tree.id: tree }
        ret, child_modes = import_git_tree(self._texts, self._mapping, "", "",
                (None, tree.id), base_inv, None, "somerevid", [],
            objects.__getitem__, (None, stat.S_IFDIR), DummyStoreUpdater(),
            self._mapping.generate_file_id)
        self.assertEquals(child_modes, {})
        self.assertEquals(2, len(ret))
        self.assertEquals(None, ret[0][0])
        self.assertEquals("", ret[0][1])
        self.assertEquals(None, ret[1][0])
        self.assertEquals("foo", ret[1][1])
        ie = ret[0][3]
        self.assertEquals("directory", ie.kind)
        ie = ret[1][3]
        self.assertEquals("file", ie.kind)
        self.assertEquals(True, ie.executable)

    def test_directory_converted_to_submodule(self):
        base_inv = Inventory()
        base_inv.add_path("foo", "directory")
        base_inv.add_path("foo/bar", "file")
        othertree = Blob.from_string("someotherthing")
        blob = Blob.from_string("bar")
        tree = Tree()
        tree.add("bar", 0160000, blob.id)
        objects = { tree.id: tree }
        ret, child_modes = import_git_submodule(self._texts, self._mapping, "foo", "foo",
                (tree.id, othertree.id), base_inv, base_inv.root.file_id, "somerevid", [],
                objects.__getitem__, (stat.S_IFDIR | 0755, S_IFGITLINK), DummyStoreUpdater(),
                self._mapping.generate_file_id)
        self.assertEquals(child_modes, {})
        self.assertEquals(2, len(ret))
        self.assertEquals(ret[0], ("foo/bar", None, base_inv.path2id("foo/bar"), None))
        self.assertEquals(ret[1][:3], ("foo", "foo", self._mapping.generate_file_id("foo")))
        ie = ret[1][3]
        self.assertEquals(ie.kind, "tree-reference")
